/* $Id: dyn_ip_test.c,v 1.28 2001/09/26 18:25:24 jm Exp $
 * Tunnel and route interface tests
 *
 * Dynamic hierarchial IP tunnel
 * Copyright (C) 1998-2001, Dynamics group
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation. See README and COPYING for
 * more details.
 */

#include <stdlib.h>
#include <stdio.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <string.h>
#include <unistd.h>
#include <sys/time.h>
#include <sys/types.h>

#include "util.h"
#include "dyn_ip.h"

#include "fixed_fd_zero.h"

#define TEST_INTERFACE "TEST"
#define TEST_ADDRESS "192.168.250.230"
#define LOCAL_ADDRESS "192.168.250.231"
#define TEST_NET "192.168.250.0"
#define TEST_NET_PLEN 24
#define TEST_NET_PLENS "24"
#define IP_PATH "/sbin/ip"
#define IFCONFIG_PATH "/sbin/ifconfig"

extern int opt_debug;

static void dyn_ip_tests(void)
{
	struct in_addr a, b, c;
	struct in_addr any;
	int ok = 1, i;

	any.s_addr = 0;

	printf("Tunnel and route interface routine tests\n"
	       "(this test must be run with root privileges)\n\n");


	inet_aton(TEST_ADDRESS, &a);
	inet_aton(LOCAL_ADDRESS, &b);
	inet_aton(TEST_NET, &c);

	i = 0;
	printf("Test case M00: Check IPIP and GRE encapsulation support\n");
	if (dyn_ip_check_ipip() < 0) {
		i = 1;
		printf("No kernel support for IPIP\n");
	}
	if (dyn_ip_check_gre() < 0) {
		i = 1;
		printf("No kernel support for GRE\n");
	}
	if (i == 1) {
		ok = 0;
		printf("M00: failed!\n\n");
	} else
		printf("M00: OK\n\n");

	printf("Test case M01: Add a tunnel\n");
	printf("The tunnel should not exist already - ifconfig:\n");
	system(IFCONFIG_PATH " " TEST_INTERFACE);
	printf("dyn_ip_tunnel_add(" TEST_INTERFACE ", " TEST_ADDRESS ")\n");
	if (dyn_ip_tunnel_add(TEST_INTERFACE, a, any) != 0) {
		fprintf(stderr, "M01: failed!\n\n");
		ok = 0;
	} else {
		printf("Tunnel should have been added - ip tu:\n");
		system(IP_PATH " tunnel show " TEST_INTERFACE);
		printf("but it should not be up: ifconfig | grep "
		       TEST_INTERFACE "\n");
		system(IFCONFIG_PATH " | grep " TEST_INTERFACE);
		printf("M01 OK\n\n");
	}


	printf("Test case M02: Set test interface up\n");
	printf("dyn_ip_link_set_dev_up(" TEST_INTERFACE ")\n");
	if (dyn_ip_link_set_dev_up(TEST_INTERFACE) != 0) {
		fprintf(stderr, "M02: failed!\n\n");
		ok = 0;
	} else {
		printf("Tunnel should now be up - ifconfig:\n");
		system(IFCONFIG_PATH " " TEST_INTERFACE);
		printf("M02 OK\n\n");
	}


	printf("Test case M03: Add local address to tunnel\n");
	printf("Tunnel should not have the local address (yet) - ifconfig:\n");
	system(IFCONFIG_PATH " " TEST_INTERFACE " | grep addr");
	printf("dyn_ip_addr_add(" TEST_INTERFACE ", " LOCAL_ADDRESS ")\n");
	if (dyn_ip_addr_add(TEST_INTERFACE, b) != 0) {
		fprintf(stderr, "M03: failed!\n\n");
		ok = 0;
	} else {
		printf("Tunnel should now have a local address - ifconfig:\n");
		system(IFCONFIG_PATH " " TEST_INTERFACE " | grep addr");
		printf("M03 OK\n\n");
	}


	printf("Test case M04: Add default route to routing table\n");
	printf("No routing info should be found yet - ip:\n");
	system(IP_PATH " route show table 99");
	printf("dyn_ip_route_add_table(" TEST_INTERFACE ", 99, NULL)\n");
	if (dyn_ip_route_add_table(TEST_INTERFACE, 99, NULL) != 0) {
		fprintf(stderr, "M04: failed!\n\n");
		ok = 0;
	} else {
		printf("Default route should have been added - ip:\n");
		system(IP_PATH " route show table 99");
		printf("M04 OK\n\n");
	}


	printf("Test case M05: Add rule for source routing\n");
	printf("No rule in table 99 should be found yet - ip:\n");
	system(IP_PATH " rule list");
	printf("dyn_ip_rule_add_table(" TEST_ADDRESS ", 99," TEST_INTERFACE
	       ")\n");
	if (dyn_ip_rule_add_table(&a, NULL, 99, TEST_INTERFACE) != 0) {
		fprintf(stderr, "M05: failed!\n\n");
		ok = 0;
	} else {
		printf("Rule should have been added to table 99 - ip:\n");
		system(IP_PATH " rule list");
		printf("M05 OK\n\n");
	}


	printf("Test case M06: Remove the previously added rule\n");
	printf("dyn_ip_rule_del_table(" TEST_ADDRESS ", 99, " TEST_INTERFACE
	       ")\n");
	if (dyn_ip_rule_del_table(&a, NULL, 99, TEST_INTERFACE) != 0) {
		fprintf(stderr, "M06: failed!\n\n");
		ok = 0;
	} else {
		printf("Rule should have been removed from table 99 - ip:\n");
		system(IP_PATH " rule list");
		printf("M06 OK\n\n");
	}


	printf("Test case M07: Add destination based route\n");
	printf("dyn_ip_route_add(" TEST_ADDRESS ", " TEST_INTERFACE ")\n");
	if (dyn_ip_route_add(a, TEST_INTERFACE) != 0) {
		fprintf(stderr, "M07: failed!\n\n");
		ok = 0;
	} else {
		printf("Route should have been added - ip:\n");
		system(IP_PATH " route");
		printf("M07 OK\n\n");
	}


	printf("Test case M07-B: Replace destination based route\n");
	printf("dyn_ip_route_replace(" TEST_ADDRESS ",eth0)\n");
	if (dyn_ip_route_replace(a, "eth0") != 0) {
		fprintf(stderr, "M07-B: failed!\n\n");
		ok = 0;
	} else {
		printf("Route should have been replaced - ip:\n");
		system(IP_PATH " route");
		printf("M07-B OK\n\n");
	}

	printf("Test case M08: Remove destination based route\n");
	printf("dyn_ip_route_del(" TEST_ADDRESS ",eth0)\n");
	if (dyn_ip_route_del(a, "eth0") != 0) { 
		fprintf(stderr, "M08: failed!\n\n");
		ok = 0;
	} else {
		printf("Route should have been deleted - ip:\n");
		system(IP_PATH " route");
		printf("M08 OK\n\n");
	}


	printf("Test case M08b: Add net route\n");
	printf("dyn_ip_route_add_net(" TEST_INTERFACE ", " TEST_NET ", %i)\n",
	       TEST_NET_PLEN);
	if (dyn_ip_route_add_net(TEST_INTERFACE, &c, TEST_NET_PLEN) != 0) {
		fprintf(stderr, "M08b: failed!\n\n");
		ok = 0;
	} else {
		printf("The route entry should have been added - ip:\n");
		system(IP_PATH " route list " TEST_NET "/" TEST_NET_PLENS);
		printf("M08b OK\n\n");
	}


	printf("Test case M08c: Delete net route\n");
	printf("dyn_ip_route_del_net(" TEST_INTERFACE ", " TEST_NET ", %i)\n",
	       TEST_NET_PLEN);
	if (dyn_ip_route_del_net(TEST_INTERFACE, &c, TEST_NET_PLEN) != 0) {
		fprintf(stderr, "M08c: failed!\n\n");
		ok = 0;
	} else {
		printf("The route entry should have been removed - ip:\n");
		system(IP_PATH " route list " TEST_NET "/" TEST_NET_PLENS);
		printf("M08c OK\n\n");
	}


	printf("Test case M09: Set the device down\n");
	printf("dyn_ip_tunnel_del(" TEST_INTERFACE ")\n");
	if (dyn_ip_link_set_dev_down(TEST_INTERFACE) != 0) {
		fprintf(stderr, "M09: failed!\n\n");
		ok = 0;
	} else {
		printf("Device should be down - ifconfig:\n");
		system(IFCONFIG_PATH " " TEST_INTERFACE);
		printf("M09 OK\n\n");
	}


	printf("Test case M10: Remove the tunnel\n");
	printf("dyn_ip_tunnel_del(" TEST_INTERFACE ")\n");
	if (dyn_ip_tunnel_del(TEST_INTERFACE) != 0) {
		fprintf(stderr, "M10: failed!\n\n");
		ok = 0;
	} else {
		printf("Tunnel should have been removed - ifconfig:\n");
		system(IFCONFIG_PATH " " TEST_INTERFACE);
		printf("M10 OK\n\n");
	}


	printf("Test case M11: Add a GRE tunnel\n");
	printf("The tunnel should not exist already - ifconfig:\n");
	system(IFCONFIG_PATH " " TEST_INTERFACE);
	printf("dyn_ip_tunnel_add_gre(" TEST_INTERFACE ", " TEST_ADDRESS
	       ",1000)\n");
	if (dyn_ip_tunnel_add_gre(TEST_INTERFACE, a, any, 1000) != 0) {
		fprintf(stderr, "M11: failed!\n\n");
		ok = 0;
	} else {
		printf("Tunnel should have been added - ip tu:\n");
		system(IP_PATH " tunnel show " TEST_INTERFACE);
		printf("but it should not be up: ifconfig | grep "
		       TEST_INTERFACE "\n");
		system(IFCONFIG_PATH " | grep " TEST_INTERFACE);
		printf("M11 OK\n\n");
	}

	printf("Test case M11: Remove the GRE tunnel\n");
	printf("dyn_ip_tunnel_del(" TEST_INTERFACE ")\n");
	if (dyn_ip_tunnel_del(TEST_INTERFACE) != 0) {
		fprintf(stderr, "M11: failed!\n\n");
		ok = 0;
	} else {
		printf("Tunnel should have been removed - ifconfig:\n");
		system(IFCONFIG_PATH " " TEST_INTERFACE);
		printf("M11 OK\n\n");
	}


	if (ok == 0) {
		fprintf(stderr, "There were some errors found during "
			"testing!\n");
		if (geteuid() != 0) {
			fprintf(stderr, "This should, however, come as no "
				"suprise because you aren't root.\n");
		}
		exit(1);
	}

	printf("Tunnel and route interface routine tests completed "
	       "successfully\n");
}

static void dyn_ip_get(char *dst)
{
	struct in_addr addr;
	char buf[10];
	int ret;

	inet_aton(dst, &addr);
	ret = dyn_ip_route_get(addr, buf, sizeof(buf));
	if (ret != 0) {
		fprintf(stderr, "dyn_ip_route_get failed => %i\n", ret);
		exit(-1);
	}
	printf("%s => %s\n", dst, buf);
}

static void dyn_ip_local(char *dst)
{
	struct in_addr addr, local;
	int ret;

	inet_aton(dst, &addr);
	ret = dyn_ip_get_local_addr(&addr, &local);
	if (ret != 0) {
		fprintf(stderr, "dyn_ip_get_local_addr failed => %i\n", ret);
		exit(-1);
	}
	printf("%s => %s\n", dst, inet_ntoa(local));
}

static void dyn_ip_show(char *dev)
{
	struct in_addr addr;
	int ret;

	ret = dyn_ip_get_ifaddr(dev, &addr);
	if (ret != 0) {
		fprintf(stderr, "dyn_ip_get_ifaddr failed => %i\n", ret);
		exit(-1);
	}
	printf("%s => %s\n", dev, inet_ntoa(addr));
}

static void dyn_ip_memory(void)
{
	struct in_addr dst;
	char dev[16];
	int i, ret;

	printf("dyn_ip_memory\n");
	for (i = 0; i < 10000; i++) {
		fprintf(stderr, ".");
		ret = dyn_ip_route_get(dst, dev, sizeof(dev));
		if (ret != 0) {
			fprintf(stderr, "dyn_ip_route_get returned %i\n", ret);
			return;
		}
	}
	printf("\ndone\n");
}

static void dyn_ip_test2(void)
{
	printf("dyn_ip_test2\n\n");
	printf("Initial default route - ip:\n");
	system(IP_PATH " route show match default");

	printf("\nSaving default route\n");
	if (dyn_ip_route_save_default(NULL) != 0) {
		printf("dyn_ip_route_save_default() failed\n");
		return;
	}

	printf("\nReplacing the default route\n"
	       "If you are using this test remotely, this may be the last\n"
	       "line you will see... But, you were warned ;)\n");
	system(IP_PATH " route replace default dev lo");
	printf("Broken default route - ip:\n");
	system(IP_PATH " route show match default");

	printf("\nRestoring the saved default route\n");
	if (dyn_ip_route_restore_default(NULL) != 0) {
		printf("dyn_ip_route_restore_default() failed\n");
	}

	printf("Returned default route - ip:\n");
	system(IP_PATH " route show match default");

	printf("\nSetting default route to dev lo with own function\n");
	if (dyn_ip_route_replace_default("lo", NULL, NULL) != 0) {
		printf("dyn_ip_route_replace_default failed\n");
	}
	printf("Default route - ip:\n");
	system(IP_PATH " route show match default");

	printf("\nRestoring the saved default route\n");
	if (dyn_ip_route_restore_default(NULL) !=  0) {
		printf("dyn_ip_route_restore_default() failed\n");
	}
	printf("Final default route - ip:\n");
	system(IP_PATH " route show match default");
}

static void dyn_ip_flush(void)
{
	printf("Old routing cache:\n");
	system(IP_PATH " route show table cache");

	printf("Flushing route cache\n");
	if (dyn_ip_route_cache_flush() != 0) {
		printf("Route cache flushing failed!\n");
	}

	printf("New routing cache:\n");
	system(IP_PATH " route show table cache");
}

static void dyn_ip_default(char *dev)
{
	if (dyn_ip_route_replace_default(dev, NULL, NULL) != 0) {
		printf("Route replace failed\n");
	}
}

static void dyn_ip_host(char *addr)
{
	struct in_addr a;

	inet_aton(addr, &a);
	if (dyn_ip_route_to_host(a) != 0)
		printf("Route to host failed\n");
}

static void dyn_ip_exists(char *_addr, char *_prefix_len)
{
	struct in_addr addr;
	int prefix_len;

	inet_aton(_addr, &addr);
	prefix_len = atoi(_prefix_len);

	printf("%s/%i - %s\n", inet_ntoa(addr), prefix_len,
	       dyn_ip_route_exists(&addr, prefix_len) ? "found" : "not found");
}


static void dyn_ip_monitor(char *dev)
{
	fd_set rfds;
	int retval;
	struct current_route_info cur;
	int fd = dyn_ip_monitor_open();

	memset(&cur, 0, sizeof(cur));
	cur.known = 1;
	inet_aton("1.2.3.4", &cur.via);
	dynamics_strlcpy(cur.ifname, dev, sizeof(cur.ifname));
	cur.ifindex = dyn_ip_get_ifindex(dev);
	if (cur.ifindex < 0) {
		fprintf(stderr, "'%s' not available\n", dev);
		exit(1);
	}
	inet_aton("192.168.200.0", &cur.net);
	cur.prefix_len = 24;
	cur.ifindex_net = cur.ifindex;
	memcpy(cur.ifname_net, cur.ifname, IFNAMSIZ);

	if (fd < 0) {
		fprintf(stderr, "Could not initiate monitoring.\n");
		return;
	}

	for (;;) {
		FD_ZERO(&rfds);
		FD_SET(fd, &rfds);
		retval = select(fd + 1, &rfds, NULL, NULL, NULL);
		if (retval <= 0) {
			printf("select returned %i\n", retval);
			exit(1);
		}

		if (!FD_ISSET(fd, &rfds)) {
			printf("!FD_ISSET?\n");
			exit(1);
		}

		printf("dyn_ip_monitor_get:\n");
		retval = dyn_ip_monitor_get(fd, &cur);
		printf("ret ==> %i\n", retval);
	}
}

static void dyn_ip_addr_if(char *iface, char *addrS)
{
	struct in_addr addr;
	int ret, ifindex;

	ifindex = atoi(iface);
	inet_aton(addrS, &addr);
	printf("ifindex=%i, addr=%s\n", ifindex, inet_ntoa(addr));
	ret = dyn_ip_iface_address(ifindex, &addr);
	printf("dyn_ip_iface_address => %i\n", ret);
}

static void dyn_ip_rule_local(void)
{
	printf("Routing rules before test:\n");
	system(IP_PATH " rule show");

	printf("Adding ip rule for locally generated packets.\n");
	if (dyn_ip_rule_add_table(NULL, NULL, 10, "lo")) {
		printf("Rule adding failed\n");
	}

	printf("Routing rules:\n");
	system(IP_PATH " rule show");

	printf("Removing ip rule for locally generated packets.\n");
	if (dyn_ip_rule_del_table(NULL, NULL, 10, "lo")) {
		printf("Rule removal failed\n");
	}

	printf("Routing rules after test:\n");
	system(IP_PATH " rule show");
}

static void show_usage(void)
{
	fprintf(stderr, "Tunnel and route interface routine tests\n"
		"dyn_ip_test test = basic module tests\n"
		"dyn_ip_test test2 = test the default route saving\n"
		"   (in case of some error, this will break your routing, so "
		"be careful ;)\n"
		"dyn_ip_test dst <dst> = get the interface for "
		"destination address\n"
		"dyn_ip_test local <dst> = get a local address of the "
		"interface of destination address\n"
		"dyn_ip_test dev <interface> = show the primary local address"
		" of the given interface\n"
		"dyn_ip_test memory = repeat operations to check for memory"
		" leaks\n"
		"dyn_ip_test default <dev> = change the default route to given"
		" device\n"
		"dyn_ip_test exists <addr prefix> <prefix len> = check "
		"routing entry existance\n"
		"dyn_ip_test monitor <iface> = monitor rtnetlink\n"
		"dyn_ip_test addr_if <iface> <addr>\n"
		"dyn_ip_test rule_local = local packets using a separate "
		"table\n"
		"dyn_ip_test flush = flush the kernel route cache\n");
	exit(-1);
}

int main(int argc, char *argv[])
{
	opt_debug = 1;

	if (argc < 2)
		show_usage();

	if (strcmp(argv[1], "test") == 0) {
		dyn_ip_tests();
		return 0;
	} else if (strcmp(argv[1], "test2") == 0) {
		dyn_ip_test2();
		return 0;
	} else if (strcmp(argv[1], "dst") == 0) {
		if (argc < 3)
			show_usage();

		dyn_ip_get(argv[2]);
		return 0;
	} else if (strcmp(argv[1], "local") == 0) {
		if (argc < 3)
			show_usage();

		dyn_ip_local(argv[2]);
		return 0;
	} else if (strcmp(argv[1], "dev") == 0) {
		if (argc < 3)
			show_usage();

		dyn_ip_show(argv[2]);
		return 0;
	} else if (strcmp(argv[1], "host") == 0) {
		if (argc < 3)
			show_usage();

		dyn_ip_host(argv[2]);
		return 0;
	} else if (strcmp(argv[1], "memory") == 0) {
		dyn_ip_memory();
		return 0;
	} else if (strcmp(argv[1], "flush") == 0) {
		dyn_ip_flush();
		return 0;
	} else if (strcmp(argv[1], "exists") == 0) {
		if (argc < 4)
			show_usage();

		dyn_ip_exists(argv[2], argv[3]);
		return 0;
	} else if (strcmp(argv[1], "monitor") == 0) {
		if (argc < 3)
			show_usage();
		dyn_ip_monitor(argv[2]);
		return 0;
	} else if (strcmp(argv[1], "addr_if") == 0) {
		if (argc < 4)
			show_usage();
		dyn_ip_addr_if(argv[2], argv[3]);
		return 0;
	} else if (strcmp(argv[1], "rule_local") == 0) {
		dyn_ip_rule_local();
		return 0;
	} else if (strcmp(argv[1], "default") == 0) {
		if (argc < 3)
			show_usage();
		dyn_ip_default(argv[1]);
		return 0;
	}

	show_usage();
	return 0; /* not reached */
}
