/* $Id: netsimul.c,v 1.20 2001/09/18 15:16:11 jm Exp $
 * Test network simulator
 *
 * Dynamic hierarchial IP tunnel
 * Copyright (C) 2001, Jouni Malinen
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation. See README and COPYING for
 * more details.
 */

#include <config.h>

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <fcntl.h>
#include <net/if.h>
#include <unistd.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/ioctl.h>
#include <netinet/if_ether.h>
#include <netinet/ip.h>
#include <netinet/ip_icmp.h>
#include <netinet/udp.h>
#include <net/if_arp.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <errno.h>
#include <time.h>
#include <assert.h>

#include "message.h"
#include "msgparser.h"
#include "auth.h"
#include "util.h"
#include "agentadv.h"


#ifdef HAVE_LINUX_IF_TUN_H
#include <linux/if_tun.h>
#endif

/* couple of defines from linux/if_tun.h if they cannot be included */
#ifndef TUNSETIFF
#if 0
/* up to and including Linux 2.4.5 */
#define TUNSETIFF (('T'<< 8) | 202)
#else
/* TUNSETIFF changed in Linux 2.4.6; at least in i386 version it is now: */
#define TUNSETIFF 0x400454ca
#endif
#endif
#ifndef IFF_TUN
#define IFF_TUN		0x0001
#endif
#ifndef IFF_TAP
#define IFF_TAP		0x0002
#endif
#ifndef IFF_NO_PI
#define IFF_NO_PI	0x1000
#endif


/* netsimul uses Linux kernel's "Universal TUN/TAP device driver'; see
 * kernel source (linux/Documentation/network/tuntap.txt) for instructions on
 * how to configure TUN/TAP */

/* Command line format:
 * <seconds from start> <command>{<field1>=<value1> ...} [option1]{..} ...
 * commands:
 *   send{dsthw,srchw,proto,dev}
 *     ip{dstip,srcip,proto}
 *     icmp{type,code,checksum}
 *     udp{dstport,srcport}
 *       req{opts,lifetime,home,ha,coa,id0,id1,saveid}
 *       rep{code,lifetime,home,ha,id0,id1,useid}
 *       mhauth{spi,secret,alg}
 *       mfauth{spi,secret,alg}
 *       fhauth{spi,secret,alg}
 *       encaps{}
 *       raw{data}
 *   expect: like send; fail if expected packet is not received
 *   reject: like expect, but fail if packet is received
 *   note{txt}
 *   exec{cmd}
 *
 * <seconds from start> can also be specified as a interval from previous
 * command (e.g., +0.5 means 0.5 sec after previous command)
 *
 * (see msgsimul1.data for an example)
 */


/* Simulated network:
 * tap0: HA=192.168.250.1, hwaddr 00:00:00:00:00:01
 * tap1: HA=192.168.251.1, hwaddr 00:00:00:00:00:02
 * tap2: HA=192.168.252.1, hwaddr 00:00:00:00:00:03
 *
 * HA=192.168.25x.1
 * MN=192.168.25x.100 .. 199
 * GW=192.168.25x.254, hwaddr 00:00:00:01:0x:fe
 * broadcast=192.168.25x.255
 *
 * foreign net 192.168.150.0/24: route to tap0
 */


int fd[3];
int max_fd;


extern int opt_debug;

static __u32 nextid = 0;
static struct timeval start;


void process(int fd, const char *dev);


struct parse_data {
	char *name;
	char *next;

	char *field;
	char *value;

	int was_last;
};

int add_ip(char *buf, int left, struct parse_data *parse);


#define PARSE_BUF_SIZE 10240

static int expect_received;
static int expect_error = 0;
static int reject_error = 0;
char current_command[PARSE_BUF_SIZE];
int current_cmd_expect;
int reject_cmd;
int current_expect;
char expect_buf[PARSE_BUF_SIZE];
struct parse_data expect_parse;

static u32 prev_id[2]; /* id used in previous send/received req/reply */


static int is_space(char c)
{
	if (c == ' ' || c == '\t' || c == 13 || c == 10)
		return 1;
	return 0;
}

static int dyn_atoi(const char *c)
{
	int i;
	if (c[0] == '0' && c[1] == 'x')
		sscanf(c, "0x%x", &i);
	else
		i = atoi(c);
	return i;
}

static u32 get_id(const char *c)
{
	if (strcmp(c, "prev0") == 0)
		return prev_id[0];
	else if (strcmp(c, "prev1") == 0)
		return prev_id[1];

	return htonl(dyn_atoi(c));
}


/* Returns:
 * 0 = no more entries
 * 1 = entry available, name in parse->name */
int parse_init(struct parse_data *parse, char *buf)
{
	char *t;

	t = buf;
	while (is_space(*t)) t++;
	parse->name = t;
	while (*t != '\0' && *t != '{') t++;
	if (*t == '\0') {
		return 0;
	} else {
		*t = '\0';
		t++;
		parse->next = t;
	}
	parse->was_last = 0;
	return 1;
}

/* Returns:
 * -1 = invalid input data
 *  0 = no more entries
 *  1 = entry available in {parse->field, parse->value}
 */
int parse_get_next(struct parse_data *parse)
{
	if (parse->was_last)
		return 0;
	parse->was_last = 0;
	while (is_space(*parse->next))
		parse->next++;
	if (*parse->next == '\0')
		return -1;
	if (*parse->next == '}') {
		parse->next++;
		return 0;
	}

	parse->field = parse->next;
	while (*parse->next != '\0' && *parse->next != '=')
		parse->next++;
	if (*parse->next == '\0')
		return -1;
	*parse->next = '\0';
	parse->next++;
	while (is_space(*parse->next))
		parse->next++;
	if (*parse->next == '"') {
		parse->next++;
		parse->value = parse->next;
		while (*parse->next != '\0' && *parse->next != '"')
			parse->next++;
		if (*parse->next == '\0')
			return -1;
		*parse->next = '\0';
		parse->next++;
	} else {
		parse->value = parse->next;
		while (*parse->next != '\0' && !is_space(*parse->next) &&
		       *parse->next != '}')
			parse->next++;
		if (*parse->next == '\0')
			return -1;
		if (*parse->next == '}')
			parse->was_last = 1;
		*parse->next = '\0';
		parse->next++;
	}

	return 1;
}

struct saved_id {
	int num;
	__u32 id[2];

	struct saved_id *next;
};

struct saved_id *saved_ids = NULL;

static void saveid(int num, __u32 id[])
{
	struct saved_id *entry;
	printf("\tsaveid[%i] = %08x:%08x\n", num, id[0], id[1]);

	entry = saved_ids;
	while (entry != NULL) {
		if (entry->num == num) {
			entry->id[0] = id[0];
			entry->id[1] = id[1];
			return;
		}
		entry = entry->next;
	}

	entry = (struct saved_id *) malloc(sizeof(struct saved_id));
	if (entry == NULL) {
		fprintf(stderr, "saveid: out of memory\n");
		return;
	}
	entry->num = num;
	entry->id[0] = id[0];
	entry->id[1] = id[1];
	entry->next = saved_ids;
	saved_ids = entry;
}

static void useid(int num, __u32 id[])
{
	struct saved_id *tmp = saved_ids;

	printf("\tuseid[%i]\n", num);

	while (tmp != NULL) {
		if (tmp->num == num) {
			id[0] = tmp->id[0];
			id[1] = tmp->id[1];
			return;
		}
		tmp = tmp->next;
	}

	fprintf(stderr, "useid - saved id=%i not found\n", num);
}


int add_req(char *buf, int left, struct parse_data *parse)
{
	struct reg_req *req;

	if (sizeof(struct reg_req) > left) {
		fprintf(stderr, "add_req - buffer full\n");
		return 0;
	}

	req = (struct reg_req *) buf;
	memset(req, 0, sizeof(struct reg_req));
	req->type = REG_REQ;
	req->id[0] = htonl(time(NULL) + UNIX_NTP_DIFF);
	req->id[1] = htonl(nextid++);

	while (parse_get_next(parse) == 1) {
		if (strcmp(parse->field, "lifetime") == 0)
			req->lifetime = htons(dyn_atoi(parse->value));
		else if (strcmp(parse->field, "opts") == 0)
			req->opts = dyn_atoi(parse->value);
		else if (strcmp(parse->field, "home") == 0)
			inet_aton(parse->value, &req->home_addr);
		else if (strcmp(parse->field, "ha") == 0)
			inet_aton(parse->value, &req->ha_addr);
		else if (strcmp(parse->field, "coa") == 0)
			inet_aton(parse->value, &req->co_addr);
		else if (strcmp(parse->field, "saveid") == 0)
			saveid(dyn_atoi(parse->value), req->id);
		else if (strcmp(parse->field, "id0") == 0)
			req->id[0] = get_id(parse->value);
		else if (strcmp(parse->field, "id1") == 0)
			req->id[1] = get_id(parse->value);
		else {
			fprintf(stderr, "add_req - unknown field '%s'\n",
				parse->field);
			return 0;
		}
	}
	return sizeof(struct reg_req);
}

int add_rep(char *buf, int left, struct parse_data *parse)
{
	struct reg_rep *rep;

	if (sizeof(struct reg_rep) > left) {
		fprintf(stderr, "add_rep - buffer full\n");
		return 0;
	}

	rep = (struct reg_rep *) buf;
	memset(rep, 0, sizeof(struct reg_rep));
	rep->type = REG_REP;
	rep->id[0] = htonl(time(NULL) + UNIX_NTP_DIFF);
	rep->id[1] = htonl(nextid++);

	while (parse_get_next(parse) == 1) {
		if (strcmp(parse->field, "lifetime") == 0)
			rep->lifetime = htons(dyn_atoi(parse->value));
		else if (strcmp(parse->field, "code") == 0)
			rep->code = dyn_atoi(parse->value);
		else if (strcmp(parse->field, "home") == 0)
			inet_aton(parse->value, &rep->home_addr);
		else if (strcmp(parse->field, "ha") == 0)
			inet_aton(parse->value, &rep->ha_addr);
		else if (strcmp(parse->field, "useid") == 0)
			useid(dyn_atoi(parse->value), rep->id);
		else if (strcmp(parse->field, "id0") == 0)
			rep->id[0] = get_id(parse->value);
		else if (strcmp(parse->field, "id1") == 0)
			rep->id[1] = get_id(parse->value);
		else {
			fprintf(stderr, "add_rep - unknown field '%s'\n",
				parse->field);
			return 0;
		}
	}

	return sizeof(struct reg_rep);
}

int add_auth(int type, char *start, char *buf, int left,
	     struct parse_data *parse)
{
	char *secret = NULL;
	int alg = AUTH_ALG_MD5_PREFIX_SUFFIX;
	int spi = 0;

	if (sizeof(struct msg_auth) + MAX_SK_LEN > left) {
		fprintf(stderr, "add_auth - buffer full\n");
		return 0;
	}

	while (parse_get_next(parse) == 1) {
		if (strcmp(parse->field, "spi") == 0)
			spi = htonl(dyn_atoi(parse->value));
		else if (strcmp(parse->field, "alg") == 0)
			alg = dyn_atoi(parse->value);
		else if (strcmp(parse->field, "secret") == 0)
			secret = parse->value;
		else {
			fprintf(stderr, "add_auth - unknown field '%s'\n",
				parse->field);
			return 0;
		}
	}
	return auth_add(alg, secret == NULL ?
			(unsigned char *) "" : (unsigned char *) secret,
			secret == NULL ? 0 : strlen(secret),
			(unsigned char *) start, (struct msg_auth *) buf,
			type, spi);
}


int add_encaps(char *buf, int left)
{
	struct encaps_delivery_ext *encaps;

	if (sizeof(struct encaps_delivery_ext) > left) {
		fprintf(stderr, "add_encaps - buffer full\n");
		return 0;
	}

	encaps = (struct encaps_delivery_ext *) buf;
	encaps->type = ENCAPS_DELIVERY_EXT;
	encaps->length = 0;

	return sizeof(struct encaps_delivery_ext);
}


int add_raw(char *buf, int left, struct parse_data *parse)
{
	char *pos = buf, *rpos;
	int i, count;

	while (parse_get_next(parse) == 1) {
		if (strcmp(parse->field, "data") == 0) {
			count = strlen(parse->value) / 2;
			if (count > left) {
				fprintf(stderr, "add_raw - buffer full\n");
				return 0;
			}
			rpos = parse->value;
			for (i = 0; i < count; i++) {
				*pos = get_hex_digit(*rpos) * 16 +
					get_hex_digit(*(rpos + 1));
				pos++;
				rpos += 2;
			}
		} else {
			fprintf(stderr, "add_raw - unknown field '%s'\n",
				parse->field);
			return 0;
		}
	}

	return pos - buf;
}


#define SENDBUF 8192

int add_udp(char *buf, int left, struct parse_data *parse)
{
	struct udphdr *udp;
	char *mip_start, *pos;
	struct msg_extensions ext;
	int mip_msg = 0;

	udp = (struct udphdr *) buf;
	udp->source = htons(434);
	udp->dest = htons(434);
	udp->check = 0;

	while (parse_get_next(parse) == 1) {
		if (strcmp(parse->field, "dstport") == 0)
			udp->dest = htons(dyn_atoi(parse->value));
		else if (strcmp(parse->field, "srcport") == 0)
			udp->source = htons(dyn_atoi(parse->value));
	}

	mip_start = pos = (char *) (udp + 1);
	while (parse_init(parse, parse->next) == 1) {
		int left = pos - buf + SENDBUF;
		printf("\tudp-data: %s\n", parse->name);
		if (strcmp(parse->name, "req") == 0) {
			pos += add_req(pos, left, parse);
			mip_msg = 1;
		} else if (strcmp(parse->name, "rep") == 0) {
			pos += add_rep(pos, left, parse);
			mip_msg = 1;
		} else if (strcmp(parse->name, "mhauth") == 0)
			pos += add_auth(MH_AUTH, mip_start, pos, left, parse);
		else if (strcmp(parse->name, "mfauth") == 0)
			pos += add_auth(MF_AUTH, mip_start, pos, left, parse);
		else if (strcmp(parse->name, "fhauth") == 0)
			pos += add_auth(FH_AUTH, mip_start, pos, left, parse);
		else if (strcmp(parse->name, "encaps") == 0)
			pos += add_encaps(pos, left);
		else if (strcmp(parse->name, "raw") == 0)
			pos += add_raw(pos, left, parse);
		else
			printf("\tunknown udp data\n");
	}

	udp->len = htons(pos - buf);
	/* FIX: UDP checksum */

	memset(&ext, 0, sizeof(ext));
	if (mip_msg && parse_msg(mip_start, pos - mip_start, &ext) != 0) {
		fprintf(stderr, "\tInvalid message!\n");
	}

	if (ext.req != NULL) {
		prev_id[0] = ext.req->id[0];
		prev_id[1] = ext.req->id[1];
	}

	if (ext.rep != NULL) {
		prev_id[0] = ext.rep->id[0];
		prev_id[1] = ext.rep->id[1];
	}

	return pos - buf;
}


int add_icmp(char *buf, int left, struct parse_data *parse)
{
	struct icmphdr *icmp;
	char *pos;
	int force_checksum = -1;

	icmp = (struct icmphdr *) buf;
	memset(icmp, 0, sizeof(*icmp));
	pos = buf + sizeof(struct icmphdr);

	while (parse_get_next(parse) == 1) {
		if (strcmp(parse->field, "type") == 0)
			icmp->type = dyn_atoi(parse->value);
		else if (strcmp(parse->field, "code") == 0)
			icmp->code = dyn_atoi(parse->value);
		else if (strcmp(parse->field, "checksum") == 0)
			force_checksum = dyn_atoi(parse->value);
	}

	while (parse_init(parse, parse->next) == 1) {
		int left = pos - buf + SENDBUF;
		printf("\ticmp-data: %s\n", parse->name);
		if (strcmp(parse->name, "ip") == 0)
			pos += add_ip(pos, left, parse);
		else if (strcmp(parse->name, "raw") == 0)
			pos += add_raw(pos, left, parse);
		else
			printf("\tunknown icmp data\n");
	}

	icmp->checksum = ip_checksum((unsigned char *) icmp, pos - buf);

	return pos - buf;
}


int add_ip(char *buf, int left, struct parse_data *parse)
{
	struct iphdr *ip;
	char *pos;
	struct in_addr addr;

	ip = (struct iphdr *) buf;
	ip->ihl = 5;
	ip->version = 4;
	ip->ttl = 255;
	ip->protocol = IPPROTO_UDP;
	ip->daddr = ip->saddr = htonl((127 << 24) | 1);

	while (parse_get_next(parse) == 1) {
		if (strcmp(parse->field, "dstip") == 0) {
			inet_aton(parse->value, &addr);
			ip->daddr = addr.s_addr;
		} else if (strcmp(parse->field, "srcip") == 0) {
			inet_aton(parse->value, &addr);
			ip->saddr = addr.s_addr;
		} else if (strcmp(parse->field, "proto") == 0) {
			ip->protocol = dyn_atoi(parse->value);
		} else if (strcmp(parse->field, "ttl") == 0) {
			ip->ttl = dyn_atoi(parse->value);
		}
	}


	pos = (char *) (ip + 1);
	while (parse_init(parse, parse->next) == 1) {
		int left = pos - buf + SENDBUF;
		printf("\tip-data: %s\n", parse->name);
		if (strcmp(parse->name, "ip") == 0)
			pos += add_ip(pos, left, parse);
		else if (strcmp(parse->name, "udp") == 0)
			pos += add_udp(pos, left, parse);
		else if (strcmp(parse->name, "icmp") == 0)
			pos += add_icmp(pos, left, parse);
		else
			printf("\tunknown ip data\n");
	}

	ip->tot_len = htons(pos - buf);
	ip->check = 0;
	ip->check = ip_checksum((unsigned char *) ip, sizeof(struct iphdr));

	return pos - buf;
}

void handle_send(struct parse_data *parse)
{
	int s = fd[0];
	char buf[SENDBUF], *pos;
	int n, dsthw_set = 0;
	struct ethhdr *eth;

	printf("handling SEND\n");

	memset(buf, 0, sizeof(buf));
	eth = (struct ethhdr *) buf;
	eth->h_proto = htons(ETH_P_IP);

	while (parse_get_next(parse) == 1) {
		if (strcmp(parse->field, "srchw") == 0) {
			if (ether_atohw(parse->value, eth->h_source) < 0)
				return;
		} else if (strcmp(parse->field, "dsthw") == 0) {
			if (ether_atohw(parse->value, eth->h_dest) < 0)
				return;
			dsthw_set = 1;
		} else if (strcmp(parse->field, "proto") == 0) {
			eth->h_proto = htons(dyn_atoi(parse->value));
		} else if (strcmp(parse->field, "dev") == 0) {
			if (strcmp(parse->value, "tap0") == 0)
				s = fd[0];
			else if (strcmp(parse->value, "tap1") == 0)
				s = fd[1];
			else if (strcmp(parse->value, "tap2") == 0)
				s = fd[2];
			else {
				fprintf(stderr, "Unknown interface '%s'\n",
					parse->value);
				return;
			}
		} else {
			fprintf(stderr, "Unknown send option '%s', value "
				"'%s'\n", parse->field, parse->value);
			return;
		}
	}

	if (!dsthw_set) {
		if (s == fd[0])
			ether_atohw("00:00:00:00:00:01", eth->h_dest);
		else if (s == fd[1])
			ether_atohw("00:00:00:00:00:02", eth->h_dest);
		else if (s == fd[2])
			ether_atohw("00:00:00:00:00:03", eth->h_dest);
	}

	pos = (char *) (eth + 1);
	while (parse_init(parse, parse->next) == 1) {
		int left = pos - buf + SENDBUF;
		printf("\tsend-data: %s\n", parse->name);
		if (strcmp(parse->name, "ip") == 0)
			pos += add_ip(pos, left, parse);
		else
			printf("\tunknown send data\n");
	}



	printf("sending %i bytes\n", pos - buf);

	n = write(s, buf, pos - buf);
	if (n < 0)
		perror("write");
	else if (n != pos - buf)
		fprintf(stderr, "sent only %i/%i bytes\n", n, pos - buf);
}


void handle_expect(struct parse_data *parse)
{
	printf("handling EXPECT\n");

	if (expect_received)
		printf("\texpected packet received\n");
	else {
		printf("\texpected packet was NOT received\n");
		expect_error++;
	}

	printf("\texpected: %s\n", current_command);
}


void handle_reject(struct parse_data *parse)
{
	printf("handling REJECT\n");

	if (!expect_received)
		printf("\texpected (reject) packet not received (OK)\n");
	else {
		printf("\texpected (reject) packet was received (FAIL)\n");
		reject_error++;
	}

	printf("\texpected (reject): %s\n", current_command);
}


void handle_note(struct parse_data *parse)
{
	printf("handling NOTE\n");

	while (parse_get_next(parse) == 1) {
		if (strcmp(parse->field, "txt") == 0)
			printf("\t%s\n", parse->value);
		else {
			fprintf(stderr, "Unknown note option '%s', value "
				"'%s'\n", parse->field, parse->value);
			return;
		}
	}
	printf("\n");
}


void handle_exec(struct parse_data *parse)
{
	printf("handling EXEC\n");

	while (parse_get_next(parse) == 1) {
		if (strcmp(parse->field, "cmd") == 0) {
			printf("running(%s)\n", parse->value);
			system(parse->value);
		} else {
			fprintf(stderr, "Unknown note option '%s', value "
				"'%s'\n", parse->field, parse->value);
			return;
		}
	}
	printf("\n");
}


void handle(char *command)
{
	char *pos, *t;
	struct timeval next, tv, diff;
	fd_set rfds;
	struct parse_data parse;
	static struct timeval prev = {0, 0};

	pos = command;
	while (*pos != '\0' && !is_space(*pos)) pos++;
	if (*pos == '\0') {
		fprintf(stderr, "Empty command[%s]\n", command);
		return;
	}

	*pos = '\0';

	pos++;
	strncpy(current_command, pos, PARSE_BUF_SIZE);
	/* parse: command{params..} .. */
	parse_init(&parse, pos);
	if (strcmp(parse.name, "expect") == 0) {
		current_cmd_expect = current_expect = 1;
		expect_received = 0;
		reject_cmd = 0;
	} else 	if (strcmp(parse.name, "reject") == 0) {
		current_cmd_expect = current_expect = 1;
		expect_received = 0;
		reject_cmd = 1;
	} else
		reject_cmd = current_cmd_expect = current_expect = 0;

	tv = start;
	if (command[0] == '+') {
		command++;
		tv = prev;
	}

	t = strchr(command, '.');
	if (t == NULL) {
		next.tv_sec = tv.tv_sec + atoi(command);
		next.tv_usec = tv.tv_usec;
	} else {
		int slen;
		*t = '\0';
		next.tv_sec = tv.tv_sec + atoi(command);
		t++;
		while (*t == '0')
			t++;
		slen = strlen(t);
		if (slen > 6)
			t[6] = '\0';
		next.tv_usec = tv.tv_usec;
		switch (slen) {
		case 1: next.tv_usec += atoi(t) * 100000; break;
		case 2: next.tv_usec += atoi(t) * 10000; break;
		case 3: next.tv_usec += atoi(t) * 1000; break;
		case 4: next.tv_usec += atoi(t) * 100; break;
		case 5: next.tv_usec += atoi(t) * 10; break;
		default: next.tv_usec += atoi(t); break;
		}
		if (next.tv_usec > 999999) {
			next.tv_usec -= 1000000;
			next.tv_sec++;
		}
	}
	prev = next;

	diff.tv_sec = next.tv_sec - start.tv_sec;
	diff.tv_usec = next.tv_usec - start.tv_usec;
	if (diff.tv_usec < 0) {
		diff.tv_sec--;
		diff.tv_usec += 1000000;
	}

	printf("next event at %lu.%06lu (start + %lu.%06lu) [%s]\n",
	       next.tv_sec, next.tv_usec, diff.tv_sec, diff.tv_usec,
	       parse.name);

	for (;;) {
		gettimeofday(&tv, NULL);
		diff.tv_sec = next.tv_sec - tv.tv_sec;
		diff.tv_usec = next.tv_usec - tv.tv_usec;
		if (diff.tv_usec < 0) {
			diff.tv_usec += 1000000;
			diff.tv_sec--;
		}
		if (diff.tv_sec < 0 || (diff.tv_sec == 0 && diff.tv_usec <= 0))
			break;

		FD_ZERO(&rfds);
		FD_SET(fd[0], &rfds);
		FD_SET(fd[1], &rfds);
		FD_SET(fd[2], &rfds);
		select(max_fd, &rfds, NULL, NULL, &diff);
		if (FD_ISSET(fd[0], &rfds))
			process(fd[0], "tap0");
		if (FD_ISSET(fd[1], &rfds))
			process(fd[1], "tap1");
		if (FD_ISSET(fd[2], &rfds))
			process(fd[2], "tap2");
	}

	diff.tv_sec = tv.tv_sec - start.tv_sec;
	diff.tv_usec = tv.tv_usec - start.tv_usec;
	if (diff.tv_usec < 0) {
		diff.tv_sec--;
		diff.tv_usec += 1000000;
	}
	printf("\nrunning command at %lu.%06lu (start + %lu.%06lu)\n\n",
	       tv.tv_sec, tv.tv_usec, diff.tv_sec, diff.tv_usec);
	if (strcmp(parse.name, "send") == 0)
		handle_send(&parse);
	else if (strcmp(parse.name, "expect") == 0)
		handle_expect(&parse);
	else if (strcmp(parse.name, "reject") == 0)
		handle_reject(&parse);
	else if (strcmp(parse.name, "note") == 0)
		handle_note(&parse);
	else if (strcmp(parse.name, "exec") == 0)
		handle_exec(&parse);
	else {
		fprintf(stderr, "unknown command[%s]\n", pos);
		return;
	}
}


int tun_alloc(char *dev)
{
    struct ifreq ifr;
    int fd, err;

    if ((fd = open("/dev/net/tun", O_RDWR)) < 0) {
	    perror("open(/dev/net/tun)");
	    return -1;
    }

    memset(&ifr, 0, sizeof(ifr));
    ifr.ifr_flags = IFF_TAP | IFF_NO_PI;
    if (dev && dev[0] != '\0')
	    dynamics_strlcpy(ifr.ifr_name, dev, IFNAMSIZ);

    if ((err = ioctl(fd, TUNSETIFF, (void *) &ifr)) < 0) {
	    perror("ioctl(TUNSETIFF)");
	    close(fd);
	    return err;
    }

    if (dev)
	    dynamics_strlcpy(dev, ifr.ifr_name, IFNAMSIZ);
    return fd;
}              


void process_router_adv(char *buf, int len)
{
	struct router_adv *adv;
	struct router_adv_router *router;
	int i, left;
	unsigned char *pos;
	struct agent_adv_ext *advext;
	struct vendor_ext_header *vend_ext;
	unsigned int u;

	printf("\tRouteradv:");

	if (len < sizeof(struct router_adv)) {
		printf("\n");
		return;
	}

	adv = (struct router_adv *) buf;
	left = len - sizeof(struct router_adv);
	printf(" num_addr=%i entry_size=%i lifetime=%i",
	       adv->num_addr, adv->entry_size, ntohs(adv->lifetime));

	router = (struct router_adv_router *) (adv + 1);
	for (i = 0; i < adv->num_addr; i++) {
		if (left < sizeof(struct router_adv_router)) {
			printf("\trouter_adv_router underflow\n");
			return;
		}
		left -= sizeof(struct router_adv_router);
		printf(" router(addr=%s, pref=%i)",
		       inet_ntoa(*(struct in_addr *)&router->router_addr),
		       ntohl(router->pref_level));
		router++;
	}
	printf("\n");

	if (left == 0)
		return;

	pos = (unsigned char *) router;
	if (*pos != AGENT_ADVERTISEMENT_EXTENSION_TYPE) {
		printf("\tUnknown router advertisement extension type %i\n",
		       *pos);
		return;
	}

	while (left > 0) {
		if (*pos != ONE_BYTE_PADDING &&
		    (left < 2 || left - 2 < pos[1])) {
			printf("\tExtension underflow\n");
			return;
		}
		switch (*pos) {
		case AGENT_ADVERTISEMENT_EXTENSION_TYPE:
			if (left < sizeof(struct agent_adv_ext)) {
				printf("\tAgentadv underflow\n");
				return;
			}
			advext = (struct agent_adv_ext *) pos;
			u = ntohs(advext->opts);
			printf("\tAgentadv: type=%i length=%i seq=%i "
			       "reg_lifetime=%i opts=0x%04X",
			       advext->type, advext->length,
			       ntohs(advext->seq), ntohs(advext->reg_lifetime),
			       u);
			if (u & AGENT_ADV_REGISTRATION_REQUIRED)
				printf(" [RegReq]");
			if (u & AGENT_ADV_BUSY)
				printf(" [Busy]");
			if (u & AGENT_ADV_HOME_AGENT)
				printf(" [HA]");
			if (u & AGENT_ADV_FOREIGN_AGENT)
				printf(" [FA]");
			if (u & AGENT_ADV_MINIMAL_ENCAPSULATION)
				printf(" [MinEncaps]");
			if (u & AGENT_ADV_GRE_ENCAPSULATION)
				printf(" [GRE]");
			if (u & AGENT_ADV_VJ_HDR_COMPRESSION)
				printf(" [VJ]");
			if (u & AGENT_ADV_BIDIR_TUNNELING)
				printf(" [BiDir]");
			u &= ~(AGENT_ADV_REGISTRATION_REQUIRED |
			       AGENT_ADV_BUSY |
			       AGENT_ADV_HOME_AGENT |
			       AGENT_ADV_FOREIGN_AGENT |
			       AGENT_ADV_MINIMAL_ENCAPSULATION |
			       AGENT_ADV_GRE_ENCAPSULATION |
			       AGENT_ADV_VJ_HDR_COMPRESSION |
			       AGENT_ADV_BIDIR_TUNNELING);
			if (u)
				printf(" [Unknown: 0x%04X]", u);
			printf("\n");
			break;

		case VENDOR_EXT_TYPE2:
			if (left < sizeof(struct vendor_ext_header)) {
				printf("\tVendor ext underflow\n");
				return;
			}
			vend_ext = (struct vendor_ext_header *) pos;
			printf("\tVendorExt: type=%i length=%i reserved=%i "
			       "vendor_id=%i sub_type=%i\n",
			       vend_ext->type, vend_ext->length,
			       vend_ext->reserved, ntohl(vend_ext->vendor_id),
			       ntohs(vend_ext->sub_type));
			if (ntohl(vend_ext->vendor_id) != VENDOR_ID_DYNAMICS ||
			    ntohs(vend_ext->sub_type) !=
			    VENDOR_EXT_DYNAMICS_AGENTADV) {
				printf("\tUnknown vendor extension\n");
			} else if (left >= sizeof(struct agent_adv_dynamics)) {
				struct agent_adv_dynamics *advd =
					(struct agent_adv_dynamics *) vend_ext;
				printf("\tDynamics adv: version=%i opts=0x%02X"
				       " reserved2=%i\n",
				       advd->version, advd->opts,
				       advd->reserved2);
			}
			break;

		case ONE_BYTE_PADDING:
			printf("\tOne byte padding\n");
			break;

		default:
			printf("\tUnknown extension type %i\n", *pos);
			break;
		}

		if (*pos == ONE_BYTE_PADDING) {
			left--;
			pos++;
		} else {
			left -= 2 + pos[1];
			pos += 2 + pos[1];
		}

	}
}

void process_icmp(char *buf, int len)
{
	struct icmphdr *icmp;

	if (len < 4)
		return;

	icmp = (struct icmphdr *) buf;

	printf("\tICMP: type=%i code=%i\n", icmp->type, icmp->code);

	switch (icmp->type) {
	case ICMP_ECHOREPLY:
		printf("\tEcho Reply\n");
		break;

	case ICMP_ECHO:
		printf("\tEcho Request\n");
		break;

	case ICMP_ROUTERADVERT:
		process_router_adv(buf, len);
		break;

	default:
		printf("\tUnknown ICMP type.\n");
		break;
	}
}


void check_auth(unsigned char *start, struct msg_auth *auth)
{
	char *secret = NULL;
	int alg = AUTH_ALG_MD5_PREFIX_SUFFIX;
	int spi = 0;

	if (auth == NULL) {
		printf("EXPECT: missing auth ext (%s)\n", expect_parse.name);
		current_expect = 0;
		return;
	}

	while (parse_get_next(&expect_parse) == 1) {
		if (strcmp(expect_parse.field, "spi") == 0) {
			spi = htonl(dyn_atoi(expect_parse.value));
			if (spi != auth->spi) {
				printf("EXPECT: spi does not match\n");
				current_expect = 0;
				break;
			}
		} else if (strcmp(expect_parse.field, "alg") == 0)
			alg = dyn_atoi(expect_parse.value);
		else if (strcmp(expect_parse.field, "secret") == 0)
			secret = expect_parse.value;
		else {
			fprintf(stderr, "check_auth - unknown field '%s'\n",
				expect_parse.field);
			current_expect = 0;
			break;
		}
	}

	if (!auth_check(alg, secret ? secret : "", secret ? strlen(secret) : 0,
			start, auth)) {
		printf("EXPECT: authentication does not match\n");
		current_expect = 0;
	}

	if (current_expect) {
		if (parse_init(&expect_parse, expect_parse.next) == 0) {
			printf("Expected packet received.\n");
			current_expect = 0;
			expect_received = 1;
		}
	}
}


void check_no_ext(const char *title, void *ext_ptr)
{
	while (parse_get_next(&expect_parse) == 1);

	if (ext_ptr != NULL) {
		printf("EXPECT: no-%s, but packet has the extension\n",
		       title);
		current_expect = 0;
	} else {
		if (parse_init(&expect_parse, expect_parse.next) == 0) {
			printf("Expected packet received.\n");
			current_expect = 0;
			expect_received = 1;
		}
	}
}


void process_udp(char *buf, int len)
{
	struct udphdr *udp;
	struct msg_extensions ext;
	struct in_addr addr;

	if (len < sizeof(struct udphdr))
		return;

	udp = (struct udphdr *) buf;
	printf("\tUDP: srcport=%i dstport=%i len=%i check=0x%04X\n",
	       ntohs(udp->source), ntohs(udp->dest), ntohs(udp->len),
	       ntohs(udp->check));

	if (current_expect && strcmp(expect_parse.name, "udp") == 0) {
		while (parse_get_next(&expect_parse) == 1) {
			if (strcmp(expect_parse.field, "dstport") == 0) {
				if (dyn_atoi(expect_parse.value) !=
				    ntohs(udp->dest)) {
					printf("EXPECT: dstport does not "
					       "match\n");
					current_expect = 0;
					break;
				}
			} else if (strcmp(expect_parse.field, "srcport") == 0)
			{
				if (dyn_atoi(expect_parse.value) !=
				    ntohs(udp->source)) {
					printf("EXPECT: srcport does not "
					       "match\n");
					current_expect = 0;
					break;
				}
			} else {
				printf("Unknown expect field udp::%s\n",
				       expect_parse.field);
				current_expect = 0;
				break;
			}
		}

		if (current_expect) {
			if (parse_init(&expect_parse, expect_parse.next) == 0)
			{
				printf("Expected packet received.\n");
				current_expect = 0;
				expect_received = 1;
			}
		}
	}

	parse_msg(buf + sizeof(struct udphdr), len - sizeof(struct udphdr),
		  &ext);

	if (current_expect && strcmp(expect_parse.name, "rep") == 0 &&
		ext.rep != NULL) {
		while (parse_get_next(&expect_parse) == 1) {
			if (strcmp(expect_parse.field, "home") == 0) {
				inet_aton(expect_parse.value, &addr);
				if (addr.s_addr != ext.rep->home_addr.s_addr) {
					printf("EXPECT: home_addr does not "
					       "match\n");
					current_expect = 0;
					break;
				}
			} else if (strcmp(expect_parse.field, "ha") == 0) {
				inet_aton(expect_parse.value, &addr);
				if (addr.s_addr != ext.rep->ha_addr.s_addr) {
					printf("EXPECT: ha_addr does not "
					       "match\n");
					current_expect = 0;
					break;
				}
			} else if (strcmp(expect_parse.field, "lifetime") == 0)
			{
				if (dyn_atoi(expect_parse.value) !=
				    ntohs(ext.rep->lifetime)) {
					printf("EXPECT: lifetime does not "
					       "match\n");
					current_expect = 0;
					break;
				}
			} else if (strcmp(expect_parse.field, "code") == 0) {
				if (dyn_atoi(expect_parse.value) !=
				    ext.rep->code) {
					printf("EXPECT: code does not "
					       "match\n");
					current_expect = 0;
					break;
				}
			} else if (strcmp(expect_parse.field, "matchid") == 0)
			{
				struct saved_id *tmp = saved_ids;
				int num = dyn_atoi(expect_parse.value);

				while (tmp != NULL) {
					if (tmp->num == num &&
					    (ext.rep->id[0] != tmp->id[0] ||
					     ext.rep->id[1] != tmp->id[1])) {
						printf("EXPECT: id does not "
						       "match\n");
						current_expect = 0;
						break;
					}
					tmp = tmp->next;
				}
				if (!current_expect)
					break;
			} else if (strcmp(expect_parse.field, "id0") == 0) {
				if (get_id(expect_parse.value) !=
				    ext.rep->id[0]) {
					printf("EXPECT: id0 does not match\n");
					current_expect = 0;
					break;
				}
			} else if (strcmp(expect_parse.field, "id1") == 0) {
				if (get_id(expect_parse.value) !=
				    ext.rep->id[1]) {
					printf("EXPECT: id1 does not match\n");
					current_expect = 0;
					break;
				}
			} else {
				printf("Unknown expect field rep::%s\n",
				       expect_parse.field);
				current_expect = 0;
				break;
			}
		}

		if (current_expect) {
			if (parse_init(&expect_parse, expect_parse.next) == 0)
			{
				printf("Expected packet received.\n");
				current_expect = 0;
				expect_received = 1;
			}
		}
	} else if (current_expect && strcmp(expect_parse.name, "req") == 0 &&
		   ext.req != NULL) {
		while (parse_get_next(&expect_parse) == 1) {
			if (strcmp(expect_parse.field, "home") == 0) {
				inet_aton(expect_parse.value, &addr);
				if (addr.s_addr != ext.req->home_addr.s_addr) {
					printf("EXPECT: home_addr does not "
					       "match\n");
					current_expect = 0;
					break;
				}
			} else if (strcmp(expect_parse.field, "ha") == 0) {
				inet_aton(expect_parse.value, &addr);
				if (addr.s_addr != ext.req->ha_addr.s_addr) {
					printf("EXPECT: ha_addr does not "
					       "match\n");
					current_expect = 0;
					break;
				}
			} else if (strcmp(expect_parse.field, "coa") == 0) {
				inet_aton(expect_parse.value, &addr);
				if (addr.s_addr != ext.req->co_addr.s_addr) {
					printf("EXPECT: co_addr does not "
					       "match\n");
					current_expect = 0;
					break;
				}
			} else if (strcmp(expect_parse.field, "lifetime") == 0)
			{
				if (dyn_atoi(expect_parse.value) !=
				    ntohs(ext.req->lifetime)) {
					printf("EXPECT: lifetime does not "
					       "match\n");
					current_expect = 0;
					break;
				}
			} else if (strcmp(expect_parse.field, "opts") == 0) {
				if (dyn_atoi(expect_parse.value) !=
				    ext.req->opts) {
					printf("EXPECT: opts does not "
					       "match\n");
					current_expect = 0;
					break;
				}
			} else if (strcmp(expect_parse.field, "matchid") == 0)
			{
				struct saved_id *tmp = saved_ids;
				int num = dyn_atoi(expect_parse.value);

				while (tmp != NULL) {
					if (tmp->num == num &&
					    (ext.req->id[0] != tmp->id[0] ||
					     ext.req->id[1] != tmp->id[1])) {
						printf("EXPECT: id does not "
						       "match\n");
						current_expect = 0;
						break;
					}
					tmp = tmp->next;
				}
				if (!current_expect)
					break;
			} else {
				printf("Unknown expect field req::%s\n",
				       expect_parse.field);
				current_expect = 0;
				break;
			}
		}

		if (current_expect) {
			if (parse_init(&expect_parse, expect_parse.next) == 0)
			{
				printf("Expected packet received.\n");
				current_expect = 0;
				expect_received = 1;
			}
		}
	}

	while (current_expect) {
		if (strcmp(expect_parse.name, "mhauth") == 0)
			check_auth(ext.start, ext.mh_auth);
		else if (strcmp(expect_parse.name, "fhauth") == 0)
			check_auth(ext.start, ext.fh_auth);
		else if (strcmp(expect_parse.name, "mfauth") == 0)
			check_auth(ext.start, ext.mf_auth);
		else if (strcmp(expect_parse.name, "no-encaps") == 0)
			check_no_ext("encaps", ext.encaps_del);
		else if (strcmp(expect_parse.name, "no-mhauth") == 0)
			check_no_ext("mhauth", ext.mh_auth);
		else if (strcmp(expect_parse.name, "no-fhauth") == 0)
			check_no_ext("fhauth", ext.fh_auth);
		else if (strcmp(expect_parse.name, "no-mfauth") == 0)
			check_no_ext("mfauth", ext.mf_auth);
		else {
			printf("EXPECT: unknown extension '%s'\n",
			       expect_parse.name);
			break;
		}
	}

	if (ext.req != NULL) {
		prev_id[0] = ext.req->id[0];
		prev_id[1] = ext.req->id[1];
	}

	if (ext.rep != NULL) {
		prev_id[0] = ext.rep->id[0];
		prev_id[1] = ext.rep->id[1];
	}
}

void process_gre(char *buf, int len)
{
	printf("\tGRE\n");
}

void process_ip(char *buf, int len)
{
	struct iphdr *ip;
	int left;
	char *next;
	struct in_addr addr;

	if (len < sizeof(struct iphdr))
		return;

	ip = (struct iphdr *) buf;
	printf("\tIP: ttl=%i protocol=%i saddr=%s", ip->ttl, ip->protocol,
	       inet_ntoa(*(struct in_addr *)&ip->saddr));
	printf(" daddr=%s\n", inet_ntoa(*(struct in_addr *)&ip->daddr));

	if (current_expect && strcmp(expect_parse.name, "ip") == 0) {
		while (parse_get_next(&expect_parse) == 1) {
			if (strcmp(expect_parse.field, "dstip") == 0) {
				inet_aton(expect_parse.value, &addr);
				if (addr.s_addr != ip->daddr) {
					printf("EXPECT: dstip does not "
					       "match\n");
					current_expect = 0;
					break;
				}
			} else if (strcmp(expect_parse.field, "srcip") == 0) {
				inet_aton(expect_parse.value, &addr);
				if (addr.s_addr != ip->saddr) {
					printf("EXPECT: srcip does not "
					       "match\n");
					current_expect = 0;
					break;
				}
			} else if (strcmp(expect_parse.field, "proto") == 0) {
				if (dyn_atoi(expect_parse.value) !=
				    ip->protocol) {
					printf("EXPECT: proto does not "
					       "match\n");
					current_expect = 0;
					break;
				}
			} else if (strcmp(expect_parse.field, "ttl") == 0) {
				if (dyn_atoi(expect_parse.value) != ip->ttl) {
					printf("EXPECT: ttl does not "
					       "match\n");
					current_expect = 0;
					break;
				}
			} else {
				printf("Unknown expect field ip::%s\n",
				       expect_parse.field);
				current_expect = 0;
				break;
			}
		}

		if (current_expect) {
			if (parse_init(&expect_parse, expect_parse.next) == 0)
			{
				printf("Expected packet received.\n");
				current_expect = 0;
				expect_received = 1;
			}
		}
	}

	left = len - sizeof(struct iphdr);
	next = buf + sizeof(struct iphdr);

	switch (ip->protocol) {
	case IPPROTO_ICMP:
		process_icmp(next, left);
		break;
	case IPPROTO_UDP:
		process_udp(next, left);
		break;
	case IPPROTO_IPIP:
		process_ip(next, left);
		break;
	case IPPROTO_GRE:
		process_gre(next, left);
		break;
	default:
		printf("\tUnknown IP protocol.\n");
		break;
	}
}


struct arphdr_ether {
	unsigned char ar_sha[ETH_ALEN];	/* Sender hardware address. */
	unsigned char ar_sip[4];	/* Sender IP address. */
	unsigned char ar_tha[ETH_ALEN];	/* Target hardware address. */
	unsigned char ar_tip[4];	/* Target IP address. */
};

void process_arp(int s, char *buf, int len)
{
	struct ethhdr *eth;
	struct arphdr *arp;
	struct arphdr_ether *arp_eth;
	int arplen;

	arplen = sizeof(struct ethhdr) + sizeof(struct arphdr) +
		sizeof(struct arphdr_ether);
	if (len != arplen) {
		fprintf(stderr, "process_arp: len=%i != arplen=%i\n",
			len, arplen);
		return;
	}

	eth = (struct ethhdr *) buf;
	arp = (struct arphdr *) (eth + 1);
	arp_eth = (struct arphdr_ether *) (arp + 1);

	printf("\tARP: ar_hrd=%i ar_pro=%i ar_hln=%i ar_pln=%i ar_op=%i\n",
	       ntohs(arp->ar_hrd), ntohs(arp->ar_pro), arp->ar_hln,
	       arp->ar_pln, ntohs(arp->ar_op));
	printf("\tARP(ether): ar_sha=%s ar_sip=%s\n",
	       ether_hwtoa(arp_eth->ar_sha),
	       inet_ntoa(*(struct in_addr *)&arp_eth->ar_sip));
	printf("\t\tar_tha=%s ar_tip=%s\n",
	       ether_hwtoa(arp_eth->ar_tha),
	       inet_ntoa(*(struct in_addr *)&arp_eth->ar_tip));

	if (ntohs(arp->ar_hrd) != ARPHRD_ETHER ||
	    ntohs(arp->ar_pro) != ETH_P_IP ||
	    arp->ar_hln != ETH_ALEN ||
	    arp->ar_pln != 4) {
		printf("\tUnknown address type\n");
		return;
	}

	if (ntohs(arp->ar_op) == ARPOP_REQUEST && arp_eth->ar_tha[0] == 0 &&
	    arp_eth->ar_tha[1] == 0 && arp_eth->ar_tha[2] == 0 &&
	    arp_eth->ar_tha[3] == 0 && arp_eth->ar_tha[4] == 0 &&
	    arp_eth->ar_tha[5] == 0 &&
	    memcmp(arp_eth->ar_tip, arp_eth->ar_sip, 4) != 0) {
		char tmp[4];
		arp->ar_op = htons(ARPOP_REPLY);
		memcpy(eth->h_dest, eth->h_source, 6);
		ether_atohw("00:12:34:56:78:90", eth->h_source);
		memcpy(arp_eth->ar_tha, arp_eth->ar_sha, 6);
		memcpy(tmp, arp_eth->ar_sip, 4);
		memcpy(arp_eth->ar_sip, arp_eth->ar_tip, 4);
		memcpy(arp_eth->ar_tip, tmp, 4);
		ether_atohw("00:12:34:56:78:90", arp_eth->ar_sha);

		printf("\tReplying to ARP request\n");
		write(s, buf, arplen);
	}
}


static void check_recv_eth(struct parse_data *parse, const char *dev,
			   struct ethhdr *eth)
{
	while (parse_get_next(parse) == 1) {
		unsigned char tmphw[ETH_ALEN];
		if (strcmp(parse->field, "srchw") == 0) {
			if (ether_atohw(parse->value, tmphw) < 0)
				fprintf(stderr, "Invalid srchw='%s'\n",
					parse->value);
			if (memcmp(tmphw, eth->h_source, ETH_ALEN) != 0) {
				printf("EXPECT: srchw does not match\n");
				current_expect = 0;
				break;
			}
		} else if (strcmp(parse->field, "dsthw") == 0) {
			if (ether_atohw(parse->value, tmphw) < 0)
				fprintf(stderr, "Invalid dsthw='%s'\n",
					parse->value);
			if (memcmp(tmphw, eth->h_dest, ETH_ALEN) != 0) {
				printf("EXPECT: dsthw does not match\n");
				current_expect = 0;
				break;
			}
		} else if (strcmp(parse->field, "proto") == 0) {
			if (htons(dyn_atoi(parse->value)) != eth->h_proto) {
				printf("EXPECT: proto does not match\n");
				current_expect = 0;
				break;
			}
		} else if (strcmp(parse->field, "dev") == 0) {
			if (strcmp(parse->value, dev) != 0) {
				printf("EXPECT: dev does not match\n");
				current_expect = 0;
				break;
			}
		} else {
			fprintf(stderr, "Unknown expect option '%s', "
				"value '%s'\n", parse->field,
				parse->value);
			current_expect = 0;
			break;
		}
	}

	if (current_expect) {
		if (parse_init(parse, parse->next) == 0) {
			printf("Expected packet received.\n");
			current_expect = 0;
			expect_received = 1;
		}
	}
}


void process(int fd, const char *dev)
{
	unsigned char buf[2000];
	int l, res;
	struct timeval tv;
	struct ethhdr *eth;

	gettimeofday(&tv, NULL);
	printf("TIME=%i.%06i   %s", (int) tv.tv_sec, (int) tv.tv_usec,
	       ctime((time_t *) &tv.tv_sec));
	l = read(fd, buf, sizeof(buf));
	printf("Received %i bytes from %s:\n", l, dev);
	if (l < 0) {
		perror("read");
		return;
	}

	eth = (struct ethhdr *) buf;
	printf("\tEthernet: dst=%s ", ether_hwtoa(eth->h_dest));
	printf("src=%s proto=%04X\n", ether_hwtoa(eth->h_source),
	       ntohs(eth->h_proto));

	current_expect = current_cmd_expect;

	if (current_expect) {
		strncpy(expect_buf, current_command, PARSE_BUF_SIZE);
		res = parse_init(&expect_parse, expect_buf);
		assert(res == 1 &&
		       (strcmp(expect_parse.name, "expect") == 0 ||
			strcmp(expect_parse.name, "reject") == 0));
		check_recv_eth(&expect_parse, dev, eth);
	}

	switch (ntohs(eth->h_proto)) {
	case ETH_P_IP:
		process_ip((char *) (eth + 1), l - sizeof(struct ethhdr));
		break;

	case ETHERTYPE_ARP:
		process_arp(fd, (char *) eth, l);
		break;

	default:
		printf("\tUnknown Ethernet Protocol ID\n");
		break;
	}
}


int main(int argc, char *argv[])
{
	char dev[IFNAMSIZ + 1];
	FILE *f;
	char buf[PARSE_BUF_SIZE];

	opt_debug = 1;
	if (argc < 2) {
		fprintf(stderr, "netsimul <command file>\n");
		exit(-1);
	}

	f = fopen(argv[1], "r");
	if (f == NULL) {
		fprintf(stderr, "Could not open command file '%s'\n", argv[1]);
		exit(-1);
	}

	dynamics_strlcpy(dev, "tap0", sizeof(dev));
	fd[0] = tun_alloc(dev);
	if (fd[0] < 0) {
		fprintf(stderr, "Could not initialize tap0\n");
		exit(-1);
	}

	dynamics_strlcpy(dev, "tap1", sizeof(dev));
	fd[1] = tun_alloc(dev);
	if (fd[1] < 0) {
		fprintf(stderr, "Could not initialize tap1\n");
		exit(-1);
	}

	dynamics_strlcpy(dev, "tap2", sizeof(dev));
	fd[2] = tun_alloc(dev);
	if (fd[2] < 0) {
		fprintf(stderr, "Could not initialize tap2\n");
		exit(-1);
	}

	max_fd = fd[1] > fd[0] ? fd[1] : fd[0];
	if (fd[2] > max_fd)
		max_fd = fd[2];
	max_fd++;

	system("/sbin/ifconfig tap0 hw ether 00:00:00:00:00:01 "
	       "192.168.250.1 up");
	system("/sbin/ifconfig tap1 hw ether 00:00:00:00:00:02 "
	       "192.168.251.1 up");
	system("/sbin/ifconfig tap2 hw ether 00:00:00:00:00:03 "
	       "192.168.252.1 up");
	system("/sbin/ip ro add 192.168.150.0/24 dev tap0");


	gettimeofday(&start, NULL);
	printf("Simulation for command file '%s' started at %lu.%06lu\n",
	       argv[1], start.tv_sec, start.tv_usec);

	while (fgets(buf, sizeof(buf), f)) {
		char *pos = buf;
		while (is_space(*pos)) pos++;
		if (*pos == ';' || *pos == '\0')
			continue; /* skip comments and empty lines */
		handle(pos);
	}

	fclose(f);

	if (expect_error) {
		printf("FAIL: number of expected packets not received: %i\n",
		       expect_error);
	}

	if (reject_error) {
		printf("FAIL: number of rejected packets received: %i\n",
		       reject_error);
	}

	return (expect_error || reject_error);
}
