/* $Id: agentadv.c,v 1.82 2001/09/29 18:29:51 jm Exp $
 * ICMP Agent Advertisement/Solicitation routines
 *
 * Dynamic hierarchial IP tunnel
 * Copyright (C) 1998-2001, Dynamics group
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation. See README and COPYING for
 * more details.
 */

#include "config.h"
#include <stdio.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/ioctl.h>
#include <netinet/in.h>
#include <netinet/ip.h>
#include <netdb.h>
#include <netinet/ip_icmp.h>
#ifdef DYN_TARGET_LINUX
#include <netinet/udp.h>
#endif
#include <arpa/inet.h>
#include <assert.h>
#include <string.h>
#include <errno.h>
#include <net/if.h>

#include <asm/types.h>

#include <unistd.h>
#include <features.h>    /* for the glibc version number */
#if __GLIBC__ >= 2 && __GLIBC_MINOR__ >= 1
#include <netpacket/packet.h>
#include <net/ethernet.h>     /* the L2 protocols */
#include <netinet/if_ether.h>
#else
#ifdef DYN_TARGET_LINUX
#include <linux/if_packet.h>
#include <linux/if_ether.h>   /* The L2 protocols */
#endif
#endif

#include "owntypes.h"
#ifdef DYN_TARGET_LINUX
#include <linux/filter.h>
#endif

#ifndef ETHERTYPE_IP
#define ETHERTYPE_IP 0x0800 /* IP */
#endif

#include "message.h"
#include "agentadv.h"
#include "dyn_ip.h"
#include "util.h"

#include "debug.h"
#define DEBUG_FLAG 'a'

/* couple of definitions from <linux/icmp.h> if they are not available
 * from the netinet/ip_icmp.h */
#ifndef ICMP_FILTER
#define ICMP_FILTER                     1
struct icmp_filter {
        __u32           data;
};
#endif

#ifndef SOL_RAW
#define SOL_RAW 255
#endif

#ifndef SO_BINDTODEVICE
#define SO_BINDTODEVICE 25
#endif


/* RFC 2002, sec. 2.1:
 * the IP destination address of an Agent Advertisement MUST be either the
 * "all systems on this link" multicast address (224.0.0.1) [5] or the
 * "limited broadcast" address (255.255.255.255)
 */
#define AGENT_ADV_DEST_IP "255.255.255.255"
#define AGENT_SOL_DEST_IP "255.255.255.255"


static struct agentadv_setup_data data;

/**
 * set_agent_adv_data:
 * @_max_lifetime: 
 * @_highest_fa_addr: 
 * @_agent_own_addr: 
 * @_agent_adv_opt: 
 * @_own_agent_adv_opt: 
 * @_adv_lifetime: 
 * @hfa_pubkey_hash: 
 * @hfa_pubkey_hash_len: 
 * 
 * 
 */
void set_agent_adv_data(__u16 _max_lifetime, struct in_addr _highest_fa_addr,
			struct in_addr _agent_own_addr, __u16 _agent_adv_opt,
			__u8 _own_agent_adv_opt, __u16 _adv_lifetime,
			unsigned char *hfa_pubkey_hash,
			int hfa_pubkey_hash_len)
{
	data.max_lifetime = htons(_max_lifetime);
	data.adv_lifetime = htons(_adv_lifetime);
	data.highest_fa_addr = (__u32)_highest_fa_addr.s_addr;
	data.agent_own_addr = (__u32)_agent_own_addr.s_addr;
	data.agent_adv_opt = htons(_agent_adv_opt);
	data.own_agent_adv_opt = _own_agent_adv_opt;
	data.hfa_pubkey_hash = hfa_pubkey_hash;
	data.hfa_pubkey_hash_len = hfa_pubkey_hash_len;
	data.fa_nai_extension = NULL;
	data.challenge = NULL;
}

/**
 * set_agent_adv_nai:
 * @nai_ext: 
 * 
 * 
 */
void set_agent_adv_nai(struct fa_nai_ext *nai_ext)
{
	data.fa_nai_extension = nai_ext;
}

void set_agent_adv_challenge(struct challenge_ext *challenge)
{
	data.challenge = challenge;
}


#ifdef INCLUDE_IPAY
/**
 * set_agent_adv_ipay:
 * @timePrice: 
 * @bytePrice: 
 * 
 * 
 */
void set_agent_adv_ipay(__u32 timePrice, __u32 bytePrice)
{
	data.timePrice = htonl(timePrice);
	data.bytePrice = htonl(bytePrice);
}
#endif


static int user_space_filter = AGENTADV_FILTER_NONE;

static int ok_buf(char *buf)
{
	__u8 proto;
#ifdef DYN_TARGET_LINUX
	struct iphdr *_ip = (struct iphdr *) buf;
	struct router_adv *_radv =
		(struct router_adv *) (buf + sizeof(struct iphdr));
#else
	struct iphdr *_ip =
		(struct iphdr *) (buf + sizeof(struct ethhdr));
	struct router_adv *_radv =
		(struct router_adv *) (buf + sizeof(struct ethhdr)
				       + sizeof(struct iphdr));
#endif


	memcpy(&proto, &_ip->protocol, 1);
	if (proto != IPPROTO_ICMP)
		return 0;

	if (user_space_filter == AGENTADV_FILTER_ADV &&
	    _radv->type != ICMP_ROUTERADVERT)
		return 0;

	if (user_space_filter == AGENTADV_FILTER_SOL &&
	    _radv->type != ICMP_ROUTERSOLICIT)
		return 0;

	return 1;
}

#ifdef DYN_TARGET_LINUX
/* iphdr - 20 octets */
static struct sock_filter agentadv_filter[] = {
	/* A <- P[9:1] - load byte - IP protocol */
	BPF_STMT(BPF_LD+BPF_B+BPF_ABS, 9),
	/* pc += (A == IPPROTO_ICMP) ? 0 : 3 */
	BPF_JUMP(BPF_JMP+BPF_JEQ+BPF_K, IPPROTO_ICMP, 0, 3),
	/* A <- P[20:1] - load byte - ICMP type */
	BPF_STMT(BPF_LD+BPF_B+BPF_ABS, 20),
	/* pc += (A == 9) ? 0 : 1 (agentadv) */
	BPF_JUMP(BPF_JMP+BPF_JEQ+BPF_K, 9, 0, 1),
	BPF_STMT(BPF_RET+BPF_K, (__u32)-1), /* accept the packet */
	BPF_STMT(BPF_RET+BPF_K, 0) /* filter packet */
};

static struct sock_fprog agentadv_fprog = {
	sizeof(agentadv_filter) / sizeof(struct sock_filter),
	agentadv_filter
};


/* iphdr - 20 octets */
static struct sock_filter agentsol_filter[] = {
	/* A <- P[9:1] - load byte - IP protocol */
	BPF_STMT(BPF_LD+BPF_B+BPF_ABS, 9),
	/* pc += (A == IPPROTO_ICMP) ? 0 : 3 */
	BPF_JUMP(BPF_JMP+BPF_JEQ+BPF_K, IPPROTO_ICMP, 0, 3),
	/* A <- P[20:1] - load byte - ICMP type */
	BPF_STMT(BPF_LD+BPF_B+BPF_ABS, 20),
	/* pc += (A == 10) ? 0 : 1 (agentsol) */
	BPF_JUMP(BPF_JMP+BPF_JEQ+BPF_K, 10, 0, 1),
	BPF_STMT(BPF_RET+BPF_K, (__u32)-1), /* accept the packet */
	BPF_STMT(BPF_RET+BPF_K, 0) /* filter packet */
};

static struct sock_fprog agentsol_fprog = {
	sizeof(agentsol_filter) / sizeof(struct sock_filter),
	agentsol_filter
};


static struct sock_filter udp434_filter[] = {
	/* A <- P[9:1] - load byte - IP protocol */
	BPF_STMT(BPF_LD+BPF_B+BPF_ABS, 9),
	/* pc += (A == IPPROTO_ICMP) ? 0 : 3 */
	BPF_JUMP(BPF_JMP+BPF_JEQ+BPF_K, IPPROTO_UDP, 0, 3),
	/* A <- P[20:1] - load half word - dest port */
	BPF_STMT(BPF_LD+BPF_H+BPF_ABS, 22),
	/* pc += (A == 434) ? 0 : 1 (agentadv) */
	BPF_JUMP(BPF_JMP+BPF_JEQ+BPF_K, 434, 0, 1),
	BPF_STMT(BPF_RET+BPF_K, (__u32)-1), /* accept the packet */
	BPF_STMT(BPF_RET+BPF_K, 0) /* filter packet */
};

static struct sock_fprog udp434_fprog = {
	sizeof(udp434_filter) / sizeof(struct sock_filter),
	udp434_filter
};
#endif

/**
 * open_agent_icmp_adv_socket:
 * @dev: 
 * 
 * 
 * 
 * Returns: 
 */
int open_agent_icmp_adv_socket(char *dev, int filter)
{
#ifdef DYN_TARGET_LINUX
	int s, i;
	struct sockaddr_ll sa;
	struct ifreq ifr;

	memset(&ifr, 0, sizeof(ifr));
	memset(&sa, 0, sizeof(sa));

	s = socket(PF_PACKET, SOCK_DGRAM, htons(ETHERTYPE_IP));
	if (s < 0) {
		if (errno == EPERM)
			DEBUG(DEBUG_FLAG, "Must be run as root\n");
		DEBUG(DEBUG_FLAG, "open_agent_icmp_adv_socket - socket: %s\n",
		      strerror(errno));
		return -1;
	}

	if (filter == AGENTADV_FILTER_ADV &&
	    setsockopt(s, SOL_SOCKET, SO_ATTACH_FILTER,
		       &agentadv_fprog, sizeof(agentadv_fprog)) < 0) {
		DEBUG(DEBUG_FLAG, "open_agent_icmp_adv_socket - setsockopt: "
		      "%s\n", strerror(errno));
		user_space_filter = filter;
	} else if (filter == AGENTADV_FILTER_SOL &&
		   setsockopt(s, SOL_SOCKET, SO_ATTACH_FILTER,
			      &agentsol_fprog, sizeof(agentsol_fprog)) < 0) {
		DEBUG(DEBUG_FLAG, "open_agent_icmp_adv_socket - setsockopt: "
		      "%s\n", strerror(errno));
		user_space_filter = filter;
	} else if (filter == UDP_FILTER_PORT434 &&
		   setsockopt(s, SOL_SOCKET, SO_ATTACH_FILTER,
			      &udp434_fprog, sizeof(udp434_fprog)) < 0) {
		DEBUG(DEBUG_FLAG, "open_agent_icmp_adv_socket - setsockopt: "
		      "%s\n", strerror(errno));
	}

	i = strlen(dev);
	if (i >= sizeof(ifr.ifr_name)) {
		DEBUG(DEBUG_FLAG, "open_agent_icmp_adv_socket: too long "
		      "interface parameter\n");
		return -1;
	}

	memcpy(ifr.ifr_name, dev, i + 1);
	if (ioctl(s, SIOCGIFINDEX, &ifr) != 0) {
		DEBUG(DEBUG_FLAG, "open_agent_icmp_adv_socket(%s): ioctl: "
		      "%s\n", dev, strerror(errno));
		close(s);
		return -1;
	}
	sa.sll_family = AF_PACKET;
	sa.sll_ifindex = ifr.ifr_ifindex;

	if (bind(s, (struct sockaddr *)&sa, sizeof(sa)) < 0) {
		DEBUG(DEBUG_FLAG, "open_agent_icmp_adv_socket - bind: %s\n",
		      strerror(errno));
		return -1;
	}

	return s;
#endif
#ifdef DYN_TARGET_WINDOWS
	struct sockaddr_in addr;
	int s;

	/* FIX: should use something to get the real ICMP packets with ethernet
	 * header.. */
	DEBUG(DEBUG_FLAG, "ICMP adv socket as a UDP kludge\n");
        if ((s = socket(AF_INET, SOCK_DGRAM, 0)) < 0) {
		DEBUG(DEBUG_FLAG, "open_agent_icmp_socket - socket: %s\n",
		      strerror(errno));
		return -1;
        }

	memset(&addr, 0, sizeof(addr));
	addr.sin_family = AF_INET;
	addr.sin_addr.s_addr = htonl(INADDR_ANY);
	addr.sin_port = htons(4344);
	if (bind(s, (struct sockaddr *) &addr, sizeof(addr)) < 0) {
		DEBUG(DEBUG_FLAG, "bind: %s\n", strerror(errno));
		close(s);
		return -1;
	}

	return s;
#endif
}


/**
 * open_agent_icmp_socket:
 * @dev: if != NULL then the socket will be bound to a local address
 * used for the device in order to send the broadcast messages to the
 * specified device 
 * @filter:
 * 
 * Returns: 
 */
int open_agent_icmp_socket(char *dev, __u32 filter)
{
	int s;

#ifdef DYN_TARGET_LINUX
        struct protoent *proto;
	struct icmp_filter ifilter;
	struct ip_mreq imr;

	if (!(proto = getprotobyname("icmp"))) {
		DEBUG(DEBUG_FLAG,
		      "getprotobyname - unknown protocol icmp - %s\n",
		      strerror(errno));
		return -1;
        }

        if ((s = socket(AF_INET, SOCK_RAW, proto->p_proto)) < 0) {
		if (errno == EPERM)
			DEBUG(DEBUG_FLAG, "Must be run as root\n");
		DEBUG(DEBUG_FLAG, "open_agent_icmp_socket - socket: %s\n",
		      strerror(errno));
		return -1;
        }

	if (dev != NULL &&
	    setsockopt(s, SOL_SOCKET, SO_BINDTODEVICE, dev, IFNAMSIZ)) {
		DEBUG(DEBUG_FLAG, "open_agent_icmp_socket - "
		      "setsockopt(SOL_SOCKET, SO_BINDTODEVICE): %s\n",
		      strerror(errno));
		close(s);
		return -1;
	}

	/* let the kernel filter out uninteresting ICMP messages */
	ifilter.data = filter;
	if (setsockopt(s, SOL_RAW, ICMP_FILTER, &ifilter, sizeof(ifilter)) < 0)
	{
		DEBUG(DEBUG_FLAG, "setsockopt: ICMP_FILTER failed: %s\n",
		      strerror(errno));
		/* do not send error code - the filtering is done in this
		 * daemon instead */
	}

	/* join multicast groups 224.0.0.1 and 224.0.0.2 */
	memset(&imr, 0, sizeof(imr));
	if (dev != NULL) {
		if (dyn_ip_get_ifaddr(dev, &imr.imr_interface) != 0) {
			DEBUG(DEBUG_FLAG, "open_agent_icmp_socket - "
			      "could not get ifaddr for %s; trying to "
			      "continue without joining multicast groups\n",
			      dev);
			return s;
		}
	}
        inet_aton("224.0.0.1", &imr.imr_multiaddr);
        if (setsockopt(s, IPPROTO_IP, IP_ADD_MEMBERSHIP, (char *) &imr,
		       sizeof(struct ip_mreq)) < 0) {
		DEBUG(DEBUG_FLAG, "setsockopt: IP_ADD_MEMBERSHIP failed: %s\n",
		      strerror(errno));
		/* try to live without multicast */
        }

        inet_aton("224.0.0.2", &imr.imr_multiaddr);
        if (setsockopt(s, IPPROTO_IP, IP_ADD_MEMBERSHIP, (char *) &imr,
		       sizeof(struct ip_mreq)) < 0) {
		DEBUG(DEBUG_FLAG, "setsockopt: IP_ADD_MEMBERSHIP failed: %s\n",
		      strerror(errno));
		/* try to live without multicast */
        }
#endif

#ifdef DYN_TARGET_WINDOWS
	int i;

	s = socket(PF_INET, SOCK_RAW, IPPROTO_ICMP);
	if (s < 0) {
		DEBUG(DEBUG_FLAG, "socket(SOCK_RAW) failed: %s\n",
		      strerror(errno));
		return -1;
	}
	i = 1;
	if (setsockopt(s, SOL_SOCKET, SO_BROADCAST, (char *) &i, sizeof(i))
	    < 0) {
		DEBUG(DEBUG_FLAG, "socket(SOCK_RAW) failed: %s\n",
		      strerror(errno));
	}
#endif

	return s;
}



#ifdef DYN_TARGET_LINUX

/**
 * send_agent_advertisement:
 * @s: packet socket (PF_PACKET, SOCK_DGRAM)
 * @dest: unicast destination address (HW) or %NULL for broadcast message
 * @destaddr: unicast destination address (IP) or %NULL for broadcast message
 * @ifindex: interface index
 * 
 * Make the agent advertisement message and send it.
 * If @dest is %NULL, use defined broadcast or multicast destination and
 * set the TTL to one.
 *
 * Returns:
 *    0) no errors,
 *   -3) send failed
 */
int send_agent_advertisement(int s, struct sockaddr *dest,
			     struct in_addr *destaddr, int ifindex)
{
	static __u16 next_agent_adv_seq = 0; /* counter used in determining 
						the advertisement sequence 
						number in _host_ byte order */
	unsigned char buf[MAX_ADV_MSG];
	struct iphdr *ip;
	struct router_adv *radv;
	struct agent_adv_ext *ext;
	struct agent_adv_dynamics *own_ext;
	int len, i, err;
	struct sockaddr_ll bc_addr;
	struct in_addr *addr, ownaddr, dstaddr;
	char dev[IFNAMSIZ];

	DEBUG(DEBUG_FLAG, "sending agent advertisement\n");

	if (dest == NULL) {
		dest = (struct sockaddr *) &bc_addr;

		memset((char *) &bc_addr, 0, sizeof(bc_addr));
		bc_addr.sll_family = AF_PACKET;
		bc_addr.sll_protocol = htons(ETH_P_IP);
		bc_addr.sll_ifindex = ifindex;
		bc_addr.sll_hatype = htons(ARPHRD_ETHER);
		bc_addr.sll_halen = ETH_ALEN; /* FIX: add support for other L2
					       * devices */
		memset(bc_addr.sll_addr, 0xff, sizeof(bc_addr.sll_addr));
	}

	if (destaddr == NULL) {
		inet_aton(AGENT_ADV_DEST_IP, &dstaddr);
		destaddr = &dstaddr;
	}

	memset(buf, 0, sizeof(buf));
	DEBUG(DEBUG_FLAG, " * IP header, len=%i\n", sizeof(struct iphdr));
	ip = (struct iphdr *) buf;
	ip->ihl = 5;
	ip->version = 4;
	ip->ttl = dest == NULL ? 1 : 255;
	ip->protocol = IPPROTO_ICMP;
	ip->daddr = destaddr->s_addr;
	if (dyn_ip_get_ifname(ifindex, dev) < 0 ||
	    dyn_ip_get_ifaddr(dev, &ownaddr) < 0)
		DEBUG(DEBUG_FLAG, "Could not get own address for agentadv\n");
	ip->saddr = ip->saddr = ownaddr.s_addr;
	len = sizeof(struct iphdr);

	/* ICMP Router Advertisement message - see RFC 1256 and
	 * RFC 2002, chapter 2 */
	DEBUG(DEBUG_FLAG, " * header, len=%i\n", sizeof(struct router_adv));

	radv = (struct router_adv *) (ip + 1);
        radv->type = ICMP_ROUTERADVERT;
        radv->code = 16; /* no common traffic routing; could also be 0
			  * if the mobility agent handles also common
			  * traffic */
        radv->checksum = 0; /* set to zero for checksum calculation */
	radv->num_addr = 0; /* Our implementation does not advertise general
			     * routers */
	radv->entry_size = 2;
	/* maximum time that this agent advertisement is consider valid */
	radv->lifetime = data.adv_lifetime;
	len += sizeof(struct router_adv);

	ext = (struct agent_adv_ext *) (buf + len);
	ext->type = 16;

	addr = (struct in_addr *) (ext + 1);
	addr->s_addr = data.agent_own_addr;
	addr++;
	/* FIX: remove own address and use only HFA address (?) */
	if (data.highest_fa_addr == 0 ||
	    data.agent_own_addr == data.highest_fa_addr) {
		ext->length = 10;
	} else {
		ext->length = 14;
		/* Change "priority". Add highest FA address as the
                   first entry */
		addr->s_addr = data.agent_own_addr;
		addr--;
		addr->s_addr = data.highest_fa_addr;
	}

	/* Sequence Number - see RFC 2002, section 2.3.2 */
	ext->seq = htons(next_agent_adv_seq);
	if (next_agent_adv_seq < 0xffff)
		next_agent_adv_seq++;
	else
		next_agent_adv_seq = 256;

	/* maximum lifetime for tunnels */
	ext->reg_lifetime = data.max_lifetime;
	ext->opts = data.agent_adv_opt;

	DEBUG(DEBUG_FLAG, " * agentadv ext, len=%i\n",
	      radv->num_addr * sizeof(struct router_adv_router) + 2 +
	      ext->length);
	len += radv->num_addr * sizeof(struct router_adv_router) +
		2 + ext->length;

	DEBUG(DEBUG_FLAG, " * Dynamics ext, len=%i\n",
	      sizeof(struct agent_adv_dynamics));
	own_ext = (struct agent_adv_dynamics *) (buf + len);
	own_ext->type = VENDOR_EXT_TYPE2;
	own_ext->length = sizeof(struct agent_adv_dynamics) - 2;
	own_ext->reserved = 0;
	own_ext->vendor_id = htonl(VENDOR_ID_DYNAMICS);
	own_ext->sub_type = htons(VENDOR_EXT_DYNAMICS_AGENTADV);
	own_ext->version = VENDOR_EXT_VERSION;
	own_ext->opts = data.own_agent_adv_opt;
	own_ext->reserved2 = 0;
#ifdef INCLUDE_IPAY
	DEBUG(DEBUG_FLAG, "ADV: timePrice=%i, bytePrice=%i\n",
	      ntohl(data.timePrice), ntohl(data.bytePrice));
	if (own_ext->length >= sizeof(struct agent_adv_dynamics) - 2) {
		own_ext->timePrice = data.timePrice;
		own_ext->bytePrice = data.bytePrice;
	} else {
		DEBUG(DEBUG_FLAG, " * no price in adv\n");
		own_ext->timePrice = 0;
		own_ext->bytePrice = 0;
	}
#endif
	len += sizeof(struct agent_adv_dynamics);

	if (data.hfa_pubkey_hash != NULL) {
		struct msg_key *hashext =
			(struct msg_key *) (buf + len);
		init_key_extension(hashext, VENDOR_EXT_DYNAMICS_PUBKEY_HASH,
				   htonl(PUBKEY_ALG_SPI_RSA),
				   data.hfa_pubkey_hash_len);
		MOVE_UNALIGNED(MSG_KEY_DATA(hashext), data.hfa_pubkey_hash,
			       data.hfa_pubkey_hash_len);
		len += GET_KEY_EXT_LEN(hashext);
		DEBUG(DEBUG_FLAG, " * pubkey hash, len=%i\n",
		      GET_KEY_EXT_LEN(hashext));
	}

	if (data.fa_nai_extension != NULL) {
		if (len + GET_NAI_EXT_LEN(data.fa_nai_extension) > MAX_ADV_MSG)
		{
			DEBUG(DEBUG_FLAG, "send_agent_advertisement - "
			      "too small buffer for FA NAI extension\n");
			return -3;
		}
		MOVE_UNALIGNED(buf + len, data.fa_nai_extension,
			       GET_NAI_EXT_LEN(data.fa_nai_extension));
		len += GET_NAI_EXT_LEN(data.fa_nai_extension);
		DEBUG(DEBUG_FLAG, " * FA NAI, len=%i\n",
		      GET_NAI_EXT_LEN(data.fa_nai_extension));
	}

	if (data.challenge != NULL) {
		int j;
		unsigned char *c;
		if (len + GET_CHALLENGE_EXT_LEN(data.challenge) > MAX_ADV_MSG)
		{
			DEBUG(DEBUG_FLAG, "send_agent_advertisement - "
			      "too small buffer for challenge extension\n");
			return -3;
		}
		MOVE_UNALIGNED(buf + len, data.challenge,
			       GET_CHALLENGE_EXT_LEN(data.challenge));
		len += GET_CHALLENGE_EXT_LEN(data.challenge);
		DEBUG(DEBUG_FLAG, " * Challenge, len=%i: ",
		      GET_CHALLENGE_EXT_LEN(data.challenge));
		c = MSG_CHALLENGE_EXT_DATA(data.challenge);
		for (j = 0; j < GET_CHALLENGE_LEN(data.challenge); j++)
			DEBUG(DEBUG_FLAG, "%02X", *c++);
		DEBUG(DEBUG_FLAG, "\n");
	}

	DEBUG(DEBUG_FLAG, " * total len: %i\n", len);
	radv->checksum = ip_checksum((unsigned char *) radv,
				     len - sizeof(struct iphdr));

	ip->tot_len = htons(len);
	ip->check = 0;
	ip->check = ip_checksum(buf, sizeof(*ip));

        i = sendto(s, buf, len, 0, dest, sizeof(struct sockaddr_ll));
	err = errno;

	if (i < 0) {
		DEBUG(DEBUG_FLAG, "send_agent_advertisement - sendto: %s\n",
		      strerror(err));
		return -3;
	}

        if (i != len) {
		DEBUG(DEBUG_FLAG, "agent adv: wrote %d/%d\n", i, len);
		return -3;
        }

#ifdef PACKET_DUMP
	DEBUG_HEXDUMP(DEBUG_FLAG, "Send agentadv", buf, i);
#endif

	return 0;
}

#endif /* DYN_TARGET_LINUX */


/**
 * send_agent_solicitation:
 * @s:
 * 
 * Make the agent solicitation message and send it
 * An Agent Solicitation message is identical to an ICMP Router Solicitation
 * [RFC1256] with the further restriction that the IP TTL field is set to 1.
 * [RFC2002]
 *
 * Returns:
 *    0) no errors,
 *   -1) setsockopt IP_TTL failed,
 *   -2) setsockopt SO_BROADCAST failed,
 *   -3) send failed
 */
int send_agent_solicitation(int s)
{
        struct agent_sol sol;
	int len;
        int i, err;
#ifdef DYN_TARGET_LINUX
	unsigned char ttl;
#endif
	struct sockaddr_in bc_addr;

	DEBUG(DEBUG_FLAG, "sending agent solicitation\n");

	memset((char *) &bc_addr, 0, sizeof(bc_addr));
	bc_addr.sin_family = AF_INET;
	inet_aton(AGENT_SOL_DEST_IP, &bc_addr.sin_addr);

	/* FIX: agent solicitation sending does not yet work with Windows */

#ifdef DYN_TARGET_LINUX
	ttl = 1;
	if (setsockopt(s, IPPROTO_IP, IP_TTL, &ttl, 1) < 0) {
		DEBUG(DEBUG_FLAG, "setsockopt: IP_TTL failed: %s\n",
		      strerror(errno));
		return -1;
	}
#endif

	i = 1;
	if (setsockopt(s, SOL_SOCKET, SO_BROADCAST, (char *) &i, sizeof(i))
	    < 0) {
		DEBUG(DEBUG_FLAG, "setsockopt: SO_BROADCAST failed: %s\n",
		      strerror(errno));
		return -2;
	}

	/* ICMP Agent solicitation - see RFC 1256 and RFC 2002 */
        sol.type = 10;
        sol.code = 0;
        sol.checksum = htons(0); /* set to zero for checksum calculation */
	sol.reserved = htonl(0);

	len = sizeof(sol);

	sol.checksum = ip_checksum((unsigned char *) &sol, len);

        i = sendto(s, &sol, len, 0, (struct sockaddr *) &bc_addr,
		   sizeof(struct sockaddr));

	err = errno;

	if (i < 0) {
		DEBUG(DEBUG_FLAG,
		      "send_agent_solicitation - sendto: %s\n", strerror(err));
		return -3;
	}
        if (i != len) {
                DEBUG(DEBUG_FLAG, "agent sol: wrote %d/%d\n", len, i);
		return -3;
        }

	return 0;
}


#ifdef DYN_TARGET_LINUX

/**
 * check_icmp_sol:
 * @s:
 * @data:
 * @from:
 * @addr:
 * 
 * Check incoming ICMP packet. 
 *
 * Returns:
 *   -1: error in receiving the message,
 *    0: router solicitation received,
 */
int check_icmp_sol(int s, struct sockaddr_ll *from, struct in_addr *dstaddr,
		   struct in_addr *srcaddr)
{
	int len;
	unsigned int fromlen;
	unsigned char buf[MAX_ADV_MSG];
	struct agent_sol *asol;
	struct iphdr *ip;
	int i;

	/* process the packet - only reply to agent sol. messages */

	fromlen = sizeof(struct sockaddr_ll);

	len = recvfrom(s, buf, sizeof(buf), 0, (struct sockaddr *) from,
		       &fromlen);
	if (len < 0) {
		DEBUG(DEBUG_FLAG, "handle_icmp_sol - recvfrom: %s\n",
		      strerror(errno));
		return -1;
	}

	ip = (struct iphdr *) buf;
	asol = (struct agent_sol *) (buf + sizeof(struct iphdr));

	/* quick checks to skip most of the ICMP messages quickly */
	if (len < sizeof(struct iphdr) + sizeof(struct agent_sol)) return -1;
	if (asol->type != ICMP_ROUTERSOLICIT)
		return -1;
	if (asol->reserved != 0) return -1;

	srcaddr->s_addr = ip->saddr;
	DEBUG(DEBUG_FLAG, "ICMP from %s ", inet_ntoa(*srcaddr));
	dstaddr->s_addr = ip->daddr;
	DEBUG(DEBUG_FLAG, "to %s, len=%i", inet_ntoa(*dstaddr), len);
	if (len != ntohs(ip->tot_len)) {
		DEBUG(DEBUG_FLAG, " => %i", ntohs(ip->tot_len));
		if (ntohs(ip->tot_len) > len) {
			DEBUG(DEBUG_FLAG, "  message underflow!");
			return -1;
		}
		len = ntohs(ip->tot_len);
		if (len < sizeof(struct iphdr) + sizeof(struct agent_sol)) {
			DEBUG(DEBUG_FLAG, "  too short!\n");
			return -1;
		}
	}
	DEBUG(DEBUG_FLAG, "\n");
	DEBUG(DEBUG_FLAG, "\tsll_ifindex=%i sll_hatype=0x%04x sll_pkttype=%i "
	      "sll_halen=%i sll_addr=",
	      from->sll_ifindex, ntohs(from->sll_hatype),
	      from->sll_pkttype, from->sll_halen);
	for (i = 0; i < sizeof(from->sll_addr); i++) {
		if (i > 0)
			DEBUG(DEBUG_FLAG, ":");
		DEBUG(DEBUG_FLAG, "%02x", from->sll_addr[i]);
	}
	DEBUG(DEBUG_FLAG, "\n");

	if (ip_checksum((unsigned char *) asol,
			len - sizeof(struct iphdr)) != 0) {
		DEBUG(DEBUG_FLAG, "\tICMP checksum error\n");
		DEBUG_HEXDUMP(DEBUG_FLAG, "Packet dump", buf, len);
		return -1;
	}

	if (len < sizeof(struct iphdr) + sizeof(struct agent_sol)) {
		DEBUG(DEBUG_FLAG, "Packet underflow - "
			"type=%i, code=%i - len=%i, expected %i\n",
		      asol->type, asol->code, len,
		      sizeof(struct iphdr) + sizeof(struct agent_sol));
		DEBUG_HEXDUMP(DEBUG_FLAG, "Packet dump", buf, len);
		return -1;
	}

	if (asol->type != 10 || asol->code != 0 || asol->reserved != 0) {
		DEBUG(DEBUG_FLAG,
		      "\tunknown ICMP - not to us (type=%i, code=%i, "
		      "reserved=%i)\n",
		      asol->type, asol->code, ntohl(asol->reserved));
		DEBUG_HEXDUMP(DEBUG_FLAG, "Packet dump", buf, len);
		return -1;
	}

	return 0;
}


/**
 * handle_icmp_sol:
 * @s:
 *
 * Get next ICMP message and reply if it is a router solicitation
 *
 * Returns:
 *   -1: error in receiving the message,
 *    0: router solicitation received and replied (agentadv sent),
 */
int handle_icmp_sol(int s)
{
	struct sockaddr_ll from;
	struct in_addr dstaddr, srcaddr;
	int r;

	r = check_icmp_sol(s, &from, &dstaddr, &srcaddr);
	if (r)
		return r;

	/* ICMP Router Solicitation Message */
	DEBUG(DEBUG_FLAG, "\tagent solicitation => reply with agent adv\n");
	send_agent_advertisement(s, (struct sockaddr *) &from, &srcaddr,
				 from.sll_ifindex);
	return 0;
}

#endif /* DYN_TARGET_LINUX */


/**
 * handle_icmp_adv:
 * @s: file descriptor of the ICMP socket for agent advertisements
 * @msg: pointer to the start of the data buffer
 * @len: maximum length of the buffer
 * @ext: pointer for the parsed agent advertisement
 *
 * Get next ICMP message and parse agent advertisement into @ext. The
 * extensions that are not included in the advertisement are set to %NULL.
 * The contents of @ext is unspecified if function fails (i.e., returned value
 * is not 1).
 *
 * Returns:
 *   1 an agent advertisement was received and there was enough buffer space
 *     for the message,
 *   0 failed - unspecified reason,
 *   -1 failed - Network is down
 */
int handle_icmp_adv(int s, char *msg, int maxlen, struct adv_extensions *ext)
{
	struct in_addr in;
	int len, used;
	socklen_t fromlen;
	struct agent_adv_dynamics *tmp_ext;
	struct in_addr addr;
#ifdef DYN_TARGET_LINUX
	unsigned char *tmp;
#endif

	assert(msg != NULL && ext != NULL);

	memset(ext, 0, sizeof(*ext));

	/* process the packet - only accept agent adv. messages */

	fromlen = sizeof(ext->from);
	len = recvfrom(s, msg, maxlen, 0, (struct sockaddr *) &ext->from,
		       &fromlen);
	if (len < 0) {
		DEBUG(DEBUG_FLAG, "handle_icmp_adv - recvfrom: %s\n",
		      strerror(errno));
		if (errno == ENETDOWN)
			return -1;
		return 0;
	}

#ifdef PACKET_DUMP
	DEBUG_HEXDUMP(DEBUG_FLAG, "Received agentadv", msg, len);
#endif

#ifdef DYN_TARGET_LINUX
	if (len < sizeof(struct iphdr) + 
	    sizeof(struct router_adv) + sizeof(struct agent_adv_ext))
		return 0;
#else
	if (len < sizeof(struct ethhdr) + sizeof(struct iphdr) + 
	    sizeof(struct router_adv) + sizeof(struct agent_adv_ext))
		return 0;
#endif

	if (user_space_filter != AGENTADV_FILTER_NONE && !ok_buf(msg))
		return 0;

	ext->start = msg;
	ext->len = len;

#ifdef DYN_TARGET_LINUX
	ext->ip = (struct iphdr *) ext->start;

	/* Packet socket seems to give some extra bytes in certain situations.
	 * Find out the correct length from the IP header. */
	tmp = (unsigned char *) &ext->ip->tot_len;
	used = *tmp * 256 + *(tmp + 1);
	if (used > ext->len) {
		DEBUG(DEBUG_FLAG, "Dropping truncated packet (%i < %i)\n",
		      ext->len, used);
		return 0;
	} else if (used < ext->len) {
		DEBUG(DEBUG_FLAG, "Fixed packet length: %i => %i\n", ext->len,
		      used);
		ext->len = used;
	}

	DEBUG(DEBUG_FLAG, "handle_icmp_adv: ifindex=%i hatype=0x%04x "
	      "pkttype=%i halen=%i addr=",
	      ext->from.sll_ifindex, ntohs(ext->from.sll_hatype),
	      ext->from.sll_pkttype, ext->from.sll_halen);
	for (used = 0; used < sizeof(ext->from.sll_addr); used++) {
		if (used > 0)
			DEBUG(DEBUG_FLAG, ":");
		DEBUG(DEBUG_FLAG, "%02x", ext->from.sll_addr[used]);
	}
	DEBUG(DEBUG_FLAG, "\n");

	used = sizeof(struct iphdr);
#else
	ext->eth = (struct ethhdr *) ext->start;
	DEBUG(DEBUG_FLAG, "handle_icmp_adv: HW-dest   = %s\n",
	      ether_hwtoa(ext->eth->h_dest));
	DEBUG(DEBUG_FLAG, "handle_icmp_adv: HW-source = %s\n",
	      ether_hwtoa(ext->eth->h_source));
	ext->ip = (struct iphdr *) (ext->eth + 1);

	used = sizeof(struct iphdr) + sizeof(struct ethhdr);
#endif

	ext->radv = (struct router_adv *) (ext->start + used);
	used += sizeof(struct router_adv);

	if (ext->radv->code != 0 && ext->radv->code != 16)
		return 0;

	UNALIGNED_(&in.s_addr, &ext->ip->saddr);
	DEBUG(DEBUG_FLAG, "ICMP from %s, len=%i\n", inet_ntoa(in),
	      ext->len);

	if (ext->radv->entry_size != 2) {
		DEBUG(DEBUG_FLAG, "\tinvalid entry_size %i\n",
		      ext->radv->entry_size);
		return 0;
	}

	if (ip_checksum((unsigned char *) ext->radv,
			ext->len -
#ifndef DYN_TARGET_LINUX
			sizeof(struct ethhdr) -
#endif
			sizeof(struct iphdr)) != 0) {
		DEBUG(DEBUG_FLAG, "\tICMP checksum error\n");
		return 0;
	}


	used += ext->radv->num_addr * sizeof(struct router_adv_router);
	ext->ext = (struct agent_adv_ext *) (ext->start + used);

	if (used + 2 > len || used + 2 + ext->ext->length > len) {
		DEBUG(DEBUG_FLAG, "\tmessage overflow - dropped\n");
		return 0;
	}

	if (ext->ext->type != AGENT_ADVERTISEMENT_EXTENSION_TYPE ||
	    ext->ext->length < 6) {
		DEBUG(DEBUG_FLAG, "\tunknown first extension type=%i, "
		      "length=%i\n", ext->ext->type, ext->ext->length);
		return 0;
	}

	tmp_ext = (struct agent_adv_dynamics *) ext->ext;
	for (;;) {
		if (tmp_ext->type == 0)
			used++; /* zero padding */
		else {
			used += 2 + tmp_ext->length;
		}
		if (used == ext->len) break;
		if (used + 1 > ext->len) {
			DEBUG(DEBUG_FLAG, "\textension overflow\n");
			return 0;
		}

		tmp_ext = (struct agent_adv_dynamics *) (ext->start + used);
		if (tmp_ext->type == 0) continue;
		if (used + 2 > ext->len ||
		    used + 2 + tmp_ext->length > ext->len) {
			DEBUG(DEBUG_FLAG, "\textension overflow - dropped\n");
			return 0;
		}
		if (ext->own_ext == NULL && tmp_ext->type == VENDOR_EXT_TYPE2
		    && tmp_ext->length >= MIN_AGENT_ADV_DYNAMICS_LEN &&
		    ntohl(tmp_ext->vendor_id) == VENDOR_ID_DYNAMICS &&
		    ntohs(tmp_ext->sub_type) == VENDOR_EXT_DYNAMICS_AGENTADV) {
			ext->own_ext = tmp_ext;
			DEBUG(DEBUG_FLAG, " * Dynamics agentadv ext\n");
			continue;
		}
		if (ext->pubkey_hash == NULL &&
		    tmp_ext->type == VENDOR_EXT_TYPE2
		    && tmp_ext->length >= sizeof(struct msg_key) &&
		    ntohl(tmp_ext->vendor_id) == VENDOR_ID_DYNAMICS &&
		    ntohs(tmp_ext->sub_type) ==
		    VENDOR_EXT_DYNAMICS_PUBKEY_HASH) {
			ext->pubkey_hash = (struct msg_key *) tmp_ext;
			DEBUG(DEBUG_FLAG, " * pubkey hash ext\n");
			continue;
		}

		if (ext->fa_nai == NULL && tmp_ext->type == VENDOR_EXT_TYPE2
		    && tmp_ext->length >= sizeof(struct fa_nai_ext) &&
		    ntohl(tmp_ext->vendor_id) == VENDOR_ID_DYNAMICS &&
		    ntohs(tmp_ext->sub_type) == VENDOR_EXT_DYNAMICS_FA_NAI) {
			unsigned char *c;
			int i;
			ext->fa_nai = (struct fa_nai_ext *) tmp_ext;
			DEBUG(DEBUG_FLAG, " * FA NAI ext: ");
			c = MSG_NAI_DATA(ext->fa_nai);
			for (i = 0; i < GET_NAI_LEN(ext->fa_nai); i++) {
				if (*c < 32 || *c > 126)
					DEBUG(DEBUG_FLAG, "<%i>", *c);
				else
					DEBUG(DEBUG_FLAG, "%c", *c);
				c++;
			}
			DEBUG(DEBUG_FLAG, "\n");
			continue;
		}

		if (ext->prefix == NULL &&
		    tmp_ext->type == PREFIX_LENGTHS_EXTENSION_TYPE) {
			ext->prefix = (struct prefix_len_ext *) tmp_ext;
			DEBUG(DEBUG_FLAG, " * prefix-lengths ext\n");
			continue;
		}

		if (ext->challenge == NULL &&
		    tmp_ext->type == AGENT_ADV_CHALLENGE_EXT) {
			int j;
			unsigned char *c;
			ext->challenge = (struct challenge_ext *) tmp_ext;
			DEBUG(DEBUG_FLAG, " * challenge ext: ");
			c = MSG_CHALLENGE_EXT_DATA(ext->challenge);
			for (j = 0; j < GET_CHALLENGE_LEN(ext->challenge); j++)
				DEBUG(DEBUG_FLAG, "%02X", *c++);
			DEBUG(DEBUG_FLAG, "\n");
			continue;
		}

		if (tmp_ext->type <= 127) {
			DEBUG(DEBUG_FLAG, "\tunknown extension (type %i) - "
			      "dropping advertisement\n", tmp_ext->type);
			return 0;
		} else {
			DEBUG(DEBUG_FLAG, "\tunknown extension (type %i) - "
			      "ignoring it, but handling the rest of the "
			      "advertisement\n", tmp_ext->type);
		}
	}

	/* ICMP Router Advertisement Message */
	UNALIGNED_(&addr.s_addr, &ext->ip->daddr);
	DEBUG(DEBUG_FLAG, "\tagent adv (sent to %s)\n", inet_ntoa(addr));

	ext->coaddrs = (ext->ext->length - 6) / 4;

#ifndef DYN_TARGET_LINUX
	assert(ext->eth);
#endif
	assert(ext->ip && ext->radv && ext->ext);

	return 1;
}
